#!/usr/bin/env python3
import json
import platform
from http.cookiejar import CookieJar
from typing import Any, Callable, Dict, List, Optional, Union

import requests
import rookiepy  # type: ignore


class GetLineAuth:
    def get_cred(self) -> Optional[str]:
        browsers: List[Callable[..., Any]] = [
            rookiepy.load,  # type: ignore # Supposed to load from any browser, but may fail
            rookiepy.firefox,  # type: ignore
            rookiepy.libre_wolf,  # type: ignore
            rookiepy.chrome,  # type: ignore
            rookiepy.chromium,  # type: ignore
            rookiepy.brave,  # type: ignore
            rookiepy.edge,  # type: ignore
            rookiepy.opera,  # type: ignore
            rookiepy.vivaldi,  # type: ignore
        ]

        if platform.system() == "Windows":
            browsers.extend(
                [
                    rookiepy.opera_gx,  # type: ignore
                    rookiepy.internet_explorer,  # type: ignore
                ]
            )
        elif platform.system() == "Darwin":
            browsers.extend(
                [
                    rookiepy.opera_gx,  # type: ignore
                    rookiepy.safari,  # type: ignore
                ]
            )

        cookies_dict = None
        cookies_jar = None
        for browser in browsers:
            try:
                cookies_dict = browser(["store.line.me"])
                cookies_jar = rookiepy.to_cookiejar(cookies_dict)  # type: ignore

                if GetLineAuth.validate_cookies(cookies_jar):
                    break

            except Exception:
                continue

        if cookies_dict is None or cookies_jar is None:
            return None

        cookies_list = ["%s=%s" % (i["name"], i["value"]) for i in cookies_dict]
        cookies = ";".join(cookies_list)

        return cookies

    @staticmethod
    def validate_cookies(cookies: Union[CookieJar, Dict[str, str]]) -> bool:
        headers = {
            "x-requested-with": "XMLHttpRequest",
        }

        params = {"text": "test"}

        response = requests.get(
            "https://store.line.me/api/custom-sticker/validate/13782/en",
            params=params,
            cookies=cookies,  # type: ignore
            headers=headers,
        )

        response_dict = json.loads(response.text)

        if response_dict["errorMessage"]:
            return False
        return True



#!/usr/bin/env python3
from __future__ import annotations

from typing import List

# References:
# https://github.com/blluv/KakaoTalkEmoticonDownloader
# https://github.com/star-39/moe-sticker-bot


class DecryptKakao:
    @staticmethod
    def generate_lfsr(key: str) -> List[int]:
        d = list(key * 2)
        seq = [0, 0, 0]

        seq[0] = 301989938
        seq[1] = 623357073
        seq[2] = -2004086252

        i = 0

        for i in range(0, 4):
            seq[0] = ord(d[i]) | (seq[0] << 8)
            seq[1] = ord(d[4 + i]) | (seq[1] << 8)
            seq[2] = ord(d[8 + i]) | (seq[2] << 8)

        seq[0] = seq[0] & 0xFFFFFFFF
        seq[1] = seq[1] & 0xFFFFFFFF
        seq[2] = seq[2] & 0xFFFFFFFF

        return seq

    @staticmethod
    def xor_byte(b: int, seq: List[int]) -> int:
        flag1 = 1
        flag2 = 0
        result = 0
        for _ in range(0, 8):
            v10 = seq[0] >> 1
            if (seq[0] << 31) & 0xFFFFFFFF:
                seq[0] = v10 ^ 0xC0000031
                v12 = seq[1] >> 1
                if (seq[1] << 31) & 0xFFFFFFFF:
                    seq[1] = (v12 | 0xC0000000) ^ 0x20000010
                    flag1 = 1
                else:
                    seq[1] = v12 & 0x3FFFFFFF
                    flag1 = 0
            else:
                seq[0] = v10
                v11 = seq[2] >> 1
                if (seq[2] << 31) & 0xFFFFFFFF:
                    seq[2] = (v11 | 0xF0000000) ^ 0x8000001
                    flag2 = 1
                else:
                    seq[2] = v11 & 0xFFFFFFF
                    flag2 = 0

            result = flag1 ^ flag2 | 2 * result
        return result ^ b

    @staticmethod
    def xor_data(data: bytes) -> bytes:
        dat = list(data)
        s = DecryptKakao.generate_lfsr("a271730728cbe141e47fd9d677e9006d")
        for i in range(0, 128):
            dat[i] = DecryptKakao.xor_byte(dat[i], s)
        return bytes(dat)



#!/usr/bin/env python3
import os
import platform
import sys
from pathlib import Path


def get_root_dir() -> Path:
    i = 0
    file_path = Path(__file__).resolve()
    while True:
        root_dir = file_path.parents[i]
        if root_dir.is_dir():
            return root_dir
        i += 1


# Directory that can read program resources
ROOT_DIR = get_root_dir()


def get_root_dir_exe() -> Path:
    if appimage_path := os.getenv("APPIMAGE"):
        return Path(appimage_path).parent

    if (
        platform.system() == "Darwin"
        and getattr(sys, "frozen", False)
        and ".app/Contents/MacOS" in ROOT_DIR.as_posix()
    ):
        return (ROOT_DIR / "../../../").resolve()

    return ROOT_DIR


# Directory that contains .exe/.app/.appimage
ROOT_DIR_EXE = get_root_dir_exe()


def check_root_dir_exe_writable() -> bool:
    if (
        not os.access(ROOT_DIR_EXE.parent, os.W_OK)
        or Path("/usr/bin/") in ROOT_DIR_EXE.parents
        or Path("/bin") in ROOT_DIR_EXE.parents
        or Path("/usr/local/bin") in ROOT_DIR_EXE.parents
        or Path("C:/Program Files") in ROOT_DIR_EXE.parents
        or Path("/Applications/") in ROOT_DIR_EXE.parents
        or "site-packages" in ROOT_DIR_EXE.as_posix()
    ):
        return False
    return True


ROOT_DIR_EXE_WRITABLE = check_root_dir_exe_writable()


def get_default_dir() -> Path:
    if ROOT_DIR_EXE_WRITABLE:
        return ROOT_DIR_EXE
    home_dir = Path.home()
    desktop_dir = home_dir / "Desktop"
    if desktop_dir.is_dir():
        return desktop_dir
    return home_dir


# Default directory for stickers_input and stickers_output
DEFAULT_DIR = get_default_dir()


def get_config_dir() -> Path:
    if platform.system() == "Windows":
        fallback_dir = Path(os.path.expandvars("%APPDATA%\\sticker-convert"))
    else:
        fallback_dir = Path.home() / ".config/sticker-convert"

    if ROOT_DIR_EXE_WRITABLE:
        return ROOT_DIR_EXE
    os.makedirs(fallback_dir, exist_ok=True)
    return fallback_dir


# Directory for saving configs
CONFIG_DIR = get_config_dir()



#!/usr/bin/env python3
from __future__ import annotations

import mmap
from decimal import ROUND_HALF_UP, Decimal
from io import BytesIO
from pathlib import Path
from typing import BinaryIO, Optional, Tuple, Union, cast

from PIL import Image, UnidentifiedImageError


class CodecInfo:
    def __init__(
        self, file: Union[Path, bytes], file_ext: Optional[str] = None
    ) -> None:
        self.file_ext: Optional[str]
        if file_ext is None and isinstance(file, Path):
            self.file_ext = CodecInfo.get_file_ext(file)
        else:
            self.file_ext = file_ext
        self.fps, self.frames, self.duration = CodecInfo.get_file_fps_frames_duration(
            file
        )
        self.codec = CodecInfo.get_file_codec(file)
        self.res = CodecInfo.get_file_res(file)
        self.is_animated = self.fps > 1

    @staticmethod
    def get_file_fps_frames_duration(
        file: Union[Path, bytes], file_ext: Optional[str] = None
    ) -> Tuple[float, int, int]:
        fps: float

        if not file_ext and isinstance(file, Path):
            file_ext = CodecInfo.get_file_ext(file)

        if file_ext == ".tgs":
            fps, frames = CodecInfo._get_file_fps_frames_tgs(file)
            if fps > 0:
                duration = int(frames / fps * 1000)
            else:
                duration = 0
        else:
            if file_ext == ".webp":
                frames, duration = CodecInfo._get_file_frames_duration_webp(file)
            elif file_ext in (".gif", ".apng", ".png"):
                frames, duration = CodecInfo._get_file_frames_duration_pillow(file)
            else:
                frames, duration = CodecInfo._get_file_frames_duration_av(file)

            if duration > 0:
                fps = frames / duration * 1000
            else:
                fps = 0

        return fps, frames, duration

    @staticmethod
    def get_file_fps(file: Union[Path, bytes], file_ext: Optional[str] = None) -> float:
        if not file_ext and isinstance(file, Path):
            file_ext = CodecInfo.get_file_ext(file)

        if file_ext == ".tgs":
            return CodecInfo._get_file_fps_tgs(file)
        if file_ext == ".webp":
            frames, duration = CodecInfo._get_file_frames_duration_webp(file)
        elif file_ext in (".gif", ".apng", ".png"):
            frames, duration = CodecInfo._get_file_frames_duration_pillow(file)
        else:
            frames, duration = CodecInfo._get_file_frames_duration_av(
                file, frames_to_iterate=10
            )

        if duration > 0:
            return frames / duration * 1000
        return 0

    @staticmethod
    def get_file_frames(
        file: Union[Path, bytes],
        file_ext: Optional[str] = None,
        check_anim: bool = False,
    ) -> int:
        # If check_anim is True, return value > 1 means the file is animated
        if not file_ext and isinstance(file, Path):
            file_ext = CodecInfo.get_file_ext(file)

        if file_ext == ".tgs":
            return CodecInfo._get_file_frames_tgs(file)
        if file_ext in (".gif", ".webp", ".png", ".apng"):
            frames, _ = CodecInfo._get_file_frames_duration_pillow(
                file, frames_only=True
            )
        else:
            if check_anim is True:
                frames_to_iterate = 2
            else:
                frames_to_iterate = None
            frames, _ = CodecInfo._get_file_frames_duration_av(
                file, frames_only=True, frames_to_iterate=frames_to_iterate
            )

        return frames

    @staticmethod
    def get_file_duration(
        file: Union[Path, bytes], file_ext: Optional[str] = None
    ) -> int:
        # Return duration in miliseconds
        if not file_ext and isinstance(file, Path):
            file_ext = CodecInfo.get_file_ext(file)

        if file_ext == ".tgs":
            fps, frames = CodecInfo._get_file_fps_frames_tgs(file)
            if fps > 0:
                duration = int(frames / fps * 1000)
            else:
                duration = 0
        elif file_ext == ".webp":
            _, duration = CodecInfo._get_file_frames_duration_webp(file)
        elif file_ext in (".gif", ".png", ".apng"):
            _, duration = CodecInfo._get_file_frames_duration_pillow(file)
        else:
            _, duration = CodecInfo._get_file_frames_duration_av(file)

        return duration

    @staticmethod
    def _get_file_fps_tgs(file: Union[Path, bytes]) -> int:
        from rlottie_python.rlottie_wrapper import LottieAnimation

        if isinstance(file, Path):
            with LottieAnimation.from_tgs(file.as_posix()) as anim:
                return anim.lottie_animation_get_framerate()
        else:
            import gzip

            with gzip.open(BytesIO(file)) as f:
                data = f.read().decode(encoding="utf-8")
            with LottieAnimation.from_data(data) as anim:
                return anim.lottie_animation_get_framerate()

    @staticmethod
    def _get_file_frames_tgs(file: Union[Path, bytes]) -> int:
        from rlottie_python.rlottie_wrapper import LottieAnimation

        if isinstance(file, Path):
            with LottieAnimation.from_tgs(file.as_posix()) as anim:
                return anim.lottie_animation_get_totalframe()
        else:
            import gzip

            with gzip.open(BytesIO(file)) as f:
                data = f.read().decode(encoding="utf-8")
            with LottieAnimation.from_data(data) as anim:
                return anim.lottie_animation_get_totalframe()

    @staticmethod
    def _get_file_fps_frames_tgs(file: Union[Path, bytes]) -> Tuple[int, int]:
        from rlottie_python.rlottie_wrapper import LottieAnimation

        if isinstance(file, Path):
            with LottieAnimation.from_tgs(file.as_posix()) as anim:
                fps = anim.lottie_animation_get_framerate()
                frames = anim.lottie_animation_get_totalframe()
        else:
            import gzip

            with gzip.open(BytesIO(file)) as f:
                data = f.read().decode(encoding="utf-8")
            with LottieAnimation.from_data(data) as anim:
                fps = anim.lottie_animation_get_framerate()
                frames = anim.lottie_animation_get_totalframe()

        return fps, frames

    @staticmethod
    def _get_file_frames_duration_pillow(
        file: Union[Path, bytes], frames_only: bool = False
    ) -> Tuple[int, int]:
        total_duration = 0

        with Image.open(file) as im:
            if "n_frames" in dir(im):
                frames = im.n_frames
                if frames_only is True:
                    return frames, 1
                for i in range(im.n_frames):
                    im.seek(i)
                    total_duration += im.info.get("duration", 1000)
                return frames, total_duration

        return 1, 0

    @staticmethod
    def _get_file_frames_duration_webp(
        file: Union[Path, bytes],
    ) -> Tuple[int, int]:
        total_duration = 0
        frames = 0

        def _open_f(file: Union[Path, bytes]) -> BinaryIO:
            if isinstance(file, Path):
                return open(file, "r+b")
            return BytesIO(file)

        with _open_f(file) as f:
            with mmap.mmap(f.fileno(), 0) as mm:
                while True:
                    anmf_pos = mm.find(b"ANMF")
                    if anmf_pos == -1:
                        break
                    mm.seek(anmf_pos + 20)
                    frame_duration_32 = mm.read(4)
                    frame_duration = frame_duration_32[:-1] + bytes(
                        int(frame_duration_32[-1]) & 0b11111100
                    )
                    total_duration += int.from_bytes(frame_duration, "little")
                    frames += 1

        if frames == 0:
            return 1, 0

        return frames, total_duration

    @staticmethod
    def _get_file_frames_duration_av(
        file: Union[Path, bytes],
        frames_to_iterate: Optional[int] = None,
        frames_only: bool = False,
    ) -> Tuple[int, int]:
        import av  # type: ignore
        from av.container.input import InputContainer  # type: ignore

        # Getting fps and frame count from metadata is not reliable
        # Example: https://github.com/laggykiller/sticker-convert/issues/114

        file_ref: Union[str, BinaryIO]
        if isinstance(file, Path):
            file_ref = file.as_posix()
        else:
            file_ref = BytesIO(file)

        with av.open(file_ref) as container:  # type: ignore
            container = cast(InputContainer, container)
            stream = container.streams.video[0]
            if container.duration:
                duration_metadata = int(
                    Decimal(container.duration / 1000).quantize(0, ROUND_HALF_UP)
                )
            else:
                duration_metadata = 0

            if frames_only is True and stream.frames > 1:
                return stream.frames, duration_metadata

            frame_count = 0
            last_frame = None
            for frame_count, frame in enumerate(container.decode(stream)):  # type: ignore
                if frames_to_iterate is not None and frame_count == frames_to_iterate:
                    break
                last_frame = frame

            if last_frame is None:
                return 0, 0

            time_base_ms = (
                last_frame.time_base.numerator / last_frame.time_base.denominator * 1000
            )
            if frame_count <= 1 or duration_metadata != 0:
                return frame_count, duration_metadata
            duration_n_minus_one = last_frame.pts * time_base_ms
            ms_per_frame = duration_n_minus_one / (frame_count - 1)
            duration = frame_count * ms_per_frame
            return frame_count, int(Decimal(duration).quantize(0, ROUND_HALF_UP))

        return 0, 0

    @staticmethod
    def get_file_codec(file: Union[Path, bytes], file_ext: Optional[str] = None) -> str:
        if not file_ext and isinstance(file, Path):
            file_ext = CodecInfo.get_file_ext(file)

        file_ref: Union[str, BinaryIO]
        if isinstance(file, Path):
            file_ref = file.as_posix()
        else:
            file_ref = BytesIO(file)

        codec = None
        animated = False
        if file_ext in (".tgs", ".lottie", ".json"):
            return file_ext.replace(".", "")
        try:
            with Image.open(file) as im:
                codec = im.format
                if "is_animated" in dir(im):
                    animated = im.is_animated
                else:
                    animated = False
        except UnidentifiedImageError:
            pass

        if codec == "PNG":
            # Unable to distinguish apng and png
            if animated:
                return "apng"
            return "png"
        if codec is not None:
            return codec.lower()

        import av  # type: ignore
        from av.error import InvalidDataError  # type: ignore

        try:
            with av.open(file_ref) as container:  # type: ignore
                return container.streams.video[0].codec_context.name.lower()
        except InvalidDataError:
            pass

        return ""

    @staticmethod
    def get_file_res(
        file: Union[Path, bytes], file_ext: Optional[str] = None
    ) -> Tuple[int, int]:
        if not file_ext and isinstance(file, Path):
            file_ext = CodecInfo.get_file_ext(file)

        if file_ext == ".tgs":
            from rlottie_python.rlottie_wrapper import LottieAnimation

            if isinstance(file, Path):
                with LottieAnimation.from_tgs(file.as_posix()) as anim:
                    width, height = anim.lottie_animation_get_size()
            else:
                import gzip

                with gzip.open(BytesIO(file)) as f:
                    data = f.read().decode(encoding="utf-8")
                with LottieAnimation.from_data(data) as anim:
                    width, height = anim.lottie_animation_get_size()
        elif file_ext in (".webp", ".png", ".apng"):
            with Image.open(file) as im:
                width = im.width
                height = im.height
        else:
            import av

            file_ref: Union[str, BinaryIO]
            if isinstance(file, Path):
                file_ref = file.as_posix()
            else:
                file_ref = BytesIO(file)

            with av.open(file_ref) as container:  # type: ignore
                stream = container.streams.video[0]
                width = stream.width
                height = stream.height

        return width, height

    @staticmethod
    def get_file_ext(file: Path) -> str:
        return Path(file).suffix.lower()

    @staticmethod
    def is_anim(file: Union[Path, bytes]) -> bool:
        if CodecInfo.get_file_frames(file, check_anim=True) > 1:
            return True
        return False



#!/usr/bin/env python3
from __future__ import annotations

from functools import partial
from typing import Any, List, Tuple

from PIL import Image, ImageDraw, ImageTk
from ttkbootstrap import Button, Canvas, Checkbutton, Entry, Frame, Label, LabelFrame, OptionMenu, Scrollbar, StringVar  # type: ignore

from sticker_convert.gui_components.frames.right_clicker import RightClicker
from sticker_convert.gui_components.gui_utils import GUIUtils
from sticker_convert.gui_components.windows.base_window import BaseWindow


class AdvancedCompressionWindow(BaseWindow):
    emoji_column_per_row = 10
    emoji_visible_rows = 5
    emoji_btns: List[Tuple[Button, ImageTk.PhotoImage]] = []

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.categories = list({entry["category"] for entry in self.gui.emoji_list})

        self.title("Advanced compression options")

        self.frame_advcomp = LabelFrame(
            self.scrollable_frame, text="Advanced compression option"
        )
        self.frame_emoji_search = LabelFrame(
            self.scrollable_frame, text="Setting default emoji"
        )
        self.frame_emoji_canvas = Frame(self.scrollable_frame)

        self.frame_advcomp.grid_columnconfigure(6, weight=1)

        cb_msg_block_adv_comp_win = partial(self.gui.cb_msg_block, parent=self)

        self.fps_help_btn = Button(
            self.frame_advcomp,
            text="?",
            width=1,
            command=lambda: cb_msg_block_adv_comp_win(self.gui.help["comp"]["fps"]),
            bootstyle="secondary",  # type: ignore
        )
        self.fps_lbl = Label(self.frame_advcomp, text="Output FPS")
        self.fps_min_lbl = Label(self.frame_advcomp, text="Min:")
        self.fps_min_entry = Entry(
            self.frame_advcomp, textvariable=self.gui.fps_min_var, width=8
        )
        self.fps_min_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)
        self.fps_max_lbl = Label(self.frame_advcomp, text="Max:")
        self.fps_max_entry = Entry(
            self.frame_advcomp, textvariable=self.gui.fps_max_var, width=8
        )
        self.fps_max_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)
        self.fps_disable_cbox = Checkbutton(
            self.frame_advcomp,
            text="X",
            variable=self.gui.fps_disable_var,
            command=self.cb_disable_fps,
            onvalue=True,
            offvalue=False,
            bootstyle="danger-round-toggle",  # type: ignore
        )

        self.res_w_help_btn = Button(
            self.frame_advcomp,
            text="?",
            width=1,
            command=lambda: cb_msg_block_adv_comp_win(self.gui.help["comp"]["res"]),
            bootstyle="secondary",  # type: ignore
        )
        self.res_w_lbl = Label(self.frame_advcomp, text="Output resolution (Width)")
        self.res_w_min_lbl = Label(self.frame_advcomp, text="Min:")
        self.res_w_min_entry = Entry(
            self.frame_advcomp, textvariable=self.gui.res_w_min_var, width=8
        )
        self.res_w_min_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)
        self.res_w_max_lbl = Label(self.frame_advcomp, text="Max:")
        self.res_w_max_entry = Entry(
            self.frame_advcomp, textvariable=self.gui.res_w_max_var, width=8
        )
        self.res_w_max_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)
        self.res_w_disable_cbox = Checkbutton(
            self.frame_advcomp,
            text="X",
            variable=self.gui.res_w_disable_var,
            command=self.cb_disable_res_w,
            onvalue=True,
            offvalue=False,
            bootstyle="danger-round-toggle",  # type: ignore
        )

        self.res_h_help_btn = Button(
            self.frame_advcomp,
            text="?",
            width=1,
            command=lambda: cb_msg_block_adv_comp_win(self.gui.help["comp"]["res"]),
            bootstyle="secondary",  # type: ignore
        )
        self.res_h_lbl = Label(self.frame_advcomp, text="Output resolution (Height)")
        self.res_h_min_lbl = Label(self.frame_advcomp, text="Min:")
        self.res_h_min_entry = Entry(
            self.frame_advcomp, textvariable=self.gui.res_h_min_var, width=8
        )
        self.res_h_min_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)
        self.res_h_max_lbl = Label(self.frame_advcomp, text="Max:")
        self.res_h_max_entry = Entry(
            self.frame_advcomp, textvariable=self.gui.res_h_max_var, width=8
        )
        self.res_h_max_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)
        self.res_h_disable_cbox = Checkbutton(
            self.frame_advcomp,
            text="X",
            variable=self.gui.res_h_disable_var,
            command=self.cb_disable_res_h,
            onvalue=True,
            offvalue=False,
            bootstyle="danger-round-toggle",  # type: ignore
        )

        self.quality_help_btn = Button(
            self.frame_advcomp,
            text="?",
            width=1,
            command=lambda: cb_msg_block_adv_comp_win(self.gui.help["comp"]["quality"]),
            bootstyle="secondary",  # type: ignore
        )
        self.quality_lbl = Label(self.frame_advcomp, text="Output quality (0-100)")
        self.quality_min_lbl = Label(self.frame_advcomp, text="Min:")
        self.quality_min_entry = Entry(
            self.frame_advcomp, textvariable=self.gui.quality_min_var, width=8
        )
        self.quality_min_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)
        self.quality_max_lbl = Label(self.frame_advcomp, text="Max:")
        self.quality_max_entry = Entry(
            self.frame_advcomp, textvariable=self.gui.quality_max_var, width=8
        )
        self.quality_max_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)
        self.quality_disable_cbox = Checkbutton(
            self.frame_advcomp,
            text="X",
            variable=self.gui.quality_disable_var,
            command=self.cb_disable_quality,
            onvalue=True,
            offvalue=False,
            bootstyle="danger-round-toggle",  # type: ignore
        )

        self.color_help_btn = Button(
            self.frame_advcomp,
            text="?",
            width=1,
            command=lambda: cb_msg_block_adv_comp_win(self.gui.help["comp"]["color"]),
            bootstyle="secondary",  # type: ignore
        )
        self.color_lbl = Label(self.frame_advcomp, text="Colors (0-256)")
        self.color_min_lbl = Label(self.frame_advcomp, text="Min:")
        self.color_min_entry = Entry(
            self.frame_advcomp, textvariable=self.gui.color_min_var, width=8
        )
        self.color_min_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)
        self.color_max_lbl = Label(self.frame_advcomp, text="Max:")
        self.color_max_entry = Entry(
            self.frame_advcomp, textvariable=self.gui.color_max_var, width=8
        )
        self.color_max_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)
        self.color_disable_cbox = Checkbutton(
            self.frame_advcomp,
            text="X",
            variable=self.gui.color_disable_var,
            command=self.cb_disable_color,
            onvalue=True,
            offvalue=False,
            bootstyle="danger-round-toggle",  # type: ignore
        )

        self.duration_help_btn = Button(
            self.frame_advcomp,
            text="?",
            width=1,
            command=lambda: cb_msg_block_adv_comp_win(
                self.gui.help["comp"]["duration"]
            ),
            bootstyle="secondary",  # type: ignore
        )
        self.duration_lbl = Label(self.frame_advcomp, text="Duration (Miliseconds)")
        self.duration_min_lbl = Label(self.frame_advcomp, text="Min:")
        self.duration_min_entry = Entry(
            self.frame_advcomp, textvariable=self.gui.duration_min_var, width=8
        )
        self.duration_min_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)
        self.duration_max_lbl = Label(self.frame_advcomp, text="Max:")
        self.duration_max_entry = Entry(
            self.frame_advcomp, textvariable=self.gui.duration_max_var, width=8
        )
        self.duration_max_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)
        self.duration_disable_cbox = Checkbutton(
            self.frame_advcomp,
            text="X",
            variable=self.gui.duration_disable_var,
            command=self.cb_disable_duration,
            onvalue=True,
            offvalue=False,
            bootstyle="danger-round-toggle",  # type: ignore
        )

        self.size_help_btn = Button(
            self.frame_advcomp,
            text="?",
            width=1,
            command=lambda: cb_msg_block_adv_comp_win(self.gui.help["comp"]["size"]),
            bootstyle="secondary",  # type: ignore
        )
        self.size_lbl = Label(self.frame_advcomp, text="Maximum file size (bytes)")
        self.img_size_max_lbl = Label(self.frame_advcomp, text="Img:")
        self.img_size_max_entry = Entry(
            self.frame_advcomp, textvariable=self.gui.img_size_max_var, width=8
        )
        self.img_size_max_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)
        self.vid_size_max_lbl = Label(self.frame_advcomp, text="Vid:")
        self.vid_size_max_entry = Entry(
            self.frame_advcomp, textvariable=self.gui.vid_size_max_var, width=8
        )
        self.vid_size_max_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)
        self.size_disable_cbox = Checkbutton(
            self.frame_advcomp,
            text="X",
            variable=self.gui.size_disable_var,
            command=self.cb_disable_size,
            onvalue=True,
            offvalue=False,
            bootstyle="danger-round-toggle",  # type: ignore
        )

        self.format_help_btn = Button(
            self.frame_advcomp,
            text="?",
            width=1,
            command=lambda: cb_msg_block_adv_comp_win(self.gui.help["comp"]["format"]),
            bootstyle="secondary",  # type: ignore
        )
        self.format_lbl = Label(self.frame_advcomp, text="File format")
        self.img_format_lbl = Label(self.frame_advcomp, text="Img:")
        self.img_format_entry = Entry(
            self.frame_advcomp, textvariable=self.gui.img_format_var, width=8
        )
        self.img_format_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)
        self.vid_format_lbl = Label(self.frame_advcomp, text="Vid:")
        self.vid_format_entry = Entry(
            self.frame_advcomp, textvariable=self.gui.vid_format_var, width=8
        )
        self.vid_format_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)

        self.power_help_btn1 = Button(
            self.frame_advcomp,
            text="?",
            width=1,
            command=lambda: cb_msg_block_adv_comp_win(
                self.gui.help["comp"]["fps_power"]
            ),
            bootstyle="secondary",  # type: ignore
        )
        self.power_lbl1 = Label(self.frame_advcomp, text="Power (Importance)")
        self.fps_power_lbl = Label(self.frame_advcomp, text="FPS:")
        self.fps_power_entry = Entry(
            self.frame_advcomp, textvariable=self.gui.fps_power_var, width=8
        )
        self.fps_power_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)
        self.res_power_lbl = Label(self.frame_advcomp, text="Res:")
        self.res_power_entry = Entry(
            self.frame_advcomp, textvariable=self.gui.res_power_var, width=8
        )
        self.res_power_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)

        self.power_help_btn2 = Button(
            self.frame_advcomp,
            text="?",
            width=1,
            command=lambda: cb_msg_block_adv_comp_win(
                self.gui.help["comp"]["fps_power"]
            ),
            bootstyle="secondary",  # type: ignore
        )
        self.power_lbl2 = Label(self.frame_advcomp, text="Power (Importance)")
        self.quality_power_lbl = Label(self.frame_advcomp, text="Quality:")
        self.quality_power_entry = Entry(
            self.frame_advcomp, textvariable=self.gui.quality_power_var, width=8
        )
        self.quality_power_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)
        self.color_power_lbl = Label(self.frame_advcomp, text="Color:")
        self.color_power_entry = Entry(
            self.frame_advcomp, textvariable=self.gui.color_power_var, width=8
        )
        self.color_power_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)

        self.fake_vid_help_btn = Button(
            self.frame_advcomp,
            text="?",
            width=1,
            command=lambda: cb_msg_block_adv_comp_win(
                self.gui.help["comp"]["fake_vid"]
            ),
            bootstyle="secondary",  # type: ignore
        )
        self.fake_vid_lbl = Label(
            self.frame_advcomp, text="Convert (faking) image to video"
        )
        self.fake_vid_cbox = Checkbutton(
            self.frame_advcomp,
            variable=self.gui.fake_vid_var,
            onvalue=True,
            offvalue=False,
            bootstyle="success-round-toggle",  # type: ignore
        )

        self.scale_filter_help_btn = Button(
            self.frame_advcomp,
            text="?",
            width=1,
            command=lambda: cb_msg_block_adv_comp_win(
                self.gui.help["comp"]["scale_filter"]
            ),
            bootstyle="secondary",  # type: ignore
        )
        self.scale_filter_lbl = Label(self.frame_advcomp, text="Scale filter")
        self.scale_filter_opt = OptionMenu(
            self.frame_advcomp,
            self.gui.scale_filter_var,
            self.gui.scale_filter_var.get(),
            "nearest",
            "box",
            "bilinear",
            "hamming",
            "bicubic",
            "lanczos",
            bootstyle="secondary",  # type: ignore
        )

        self.quantize_method_help_btn = Button(
            self.frame_advcomp,
            text="?",
            width=1,
            command=lambda: cb_msg_block_adv_comp_win(
                self.gui.help["comp"]["quantize_method"]
            ),
            bootstyle="secondary",  # type: ignore
        )
        self.quantize_method_lbl = Label(self.frame_advcomp, text="Quantize method")
        self.quantize_method_opt = OptionMenu(
            self.frame_advcomp,
            self.gui.quantize_method_var,
            self.gui.quantize_method_var.get(),
            "imagequant",
            "fastoctree",
            "none",
            bootstyle="secondary",  # type: ignore
        )

        self.cache_dir_help_btn = Button(
            self.frame_advcomp,
            text="?",
            width=1,
            command=lambda: cb_msg_block_adv_comp_win(
                self.gui.help["comp"]["cache_dir"]
            ),
            bootstyle="secondary",  # type: ignore
        )
        self.cache_dir_lbl = Label(self.frame_advcomp, text="Custom cache directory")
        self.cache_dir_entry = Entry(
            self.frame_advcomp, textvariable=self.gui.cache_dir_var, width=30
        )
        self.cache_dir_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)

        self.default_emoji_help_btn = Button(
            self.frame_advcomp,
            text="?",
            width=1,
            command=lambda: cb_msg_block_adv_comp_win(
                self.gui.help["comp"]["default_emoji"]
            ),
            bootstyle="secondary",  # type: ignore
        )
        self.default_emoji_lbl = Label(self.frame_advcomp, text="Default emoji")
        self.im: Image.Image = Image.new("RGBA", (32, 32), (255, 255, 255, 0))
        self.ph_im = ImageTk.PhotoImage(self.im)
        self.default_emoji_dsp = Label(self.frame_advcomp, image=self.ph_im)

        self.fps_help_btn.grid(column=0, row=2, sticky="w", padx=3, pady=3)
        self.fps_lbl.grid(column=1, row=2, sticky="w", padx=3, pady=3)
        self.fps_min_lbl.grid(column=2, row=2, sticky="w", padx=3, pady=3)
        self.fps_min_entry.grid(column=3, row=2, sticky="nes", padx=3, pady=3)
        self.fps_max_lbl.grid(column=4, row=2, sticky="w", padx=3, pady=3)
        self.fps_max_entry.grid(column=5, row=2, sticky="nes", padx=3, pady=3)
        self.fps_disable_cbox.grid(column=6, row=2, sticky="nes", padx=3, pady=3)

        self.res_w_help_btn.grid(column=0, row=3, sticky="w", padx=3, pady=3)
        self.res_w_lbl.grid(column=1, row=3, sticky="w", padx=3, pady=3)
        self.res_w_min_lbl.grid(column=2, row=3, sticky="w", padx=3, pady=3)
        self.res_w_min_entry.grid(column=3, row=3, sticky="nes", padx=3, pady=3)
        self.res_w_max_lbl.grid(column=4, row=3, sticky="w", padx=3, pady=3)
        self.res_w_max_entry.grid(column=5, row=3, sticky="nes", padx=3, pady=3)
        self.res_w_disable_cbox.grid(column=6, row=3, sticky="nes", padx=3, pady=3)

        self.res_h_help_btn.grid(column=0, row=4, sticky="w", padx=3, pady=3)
        self.res_h_lbl.grid(column=1, row=4, sticky="w", padx=3, pady=3)
        self.res_h_min_lbl.grid(column=2, row=4, sticky="w", padx=3, pady=3)
        self.res_h_min_entry.grid(column=3, row=4, sticky="nes", padx=3, pady=3)
        self.res_h_max_lbl.grid(column=4, row=4, sticky="w", padx=3, pady=3)
        self.res_h_max_entry.grid(column=5, row=4, sticky="nes", padx=3, pady=3)
        self.res_h_disable_cbox.grid(column=6, row=4, sticky="nes", padx=3, pady=3)

        self.quality_help_btn.grid(column=0, row=5, sticky="w", padx=3, pady=3)
        self.quality_lbl.grid(column=1, row=5, sticky="w", padx=3, pady=3)
        self.quality_min_lbl.grid(column=2, row=5, sticky="w", padx=3, pady=3)
        self.quality_min_entry.grid(column=3, row=5, sticky="nes", padx=3, pady=3)
        self.quality_max_lbl.grid(column=4, row=5, sticky="w", padx=3, pady=3)
        self.quality_max_entry.grid(column=5, row=5, sticky="nes", padx=3, pady=3)
        self.quality_disable_cbox.grid(column=6, row=5, sticky="nes", padx=3, pady=3)

        self.color_help_btn.grid(column=0, row=6, sticky="w", padx=3, pady=3)
        self.color_lbl.grid(column=1, row=6, sticky="w", padx=3, pady=3)
        self.color_min_lbl.grid(column=2, row=6, sticky="w", padx=3, pady=3)
        self.color_min_entry.grid(column=3, row=6, sticky="nes", padx=3, pady=3)
        self.color_max_lbl.grid(column=4, row=6, sticky="w", padx=3, pady=3)
        self.color_max_entry.grid(column=5, row=6, sticky="nes", padx=3, pady=3)
        self.color_disable_cbox.grid(column=6, row=6, sticky="nes", padx=3, pady=3)

        self.duration_help_btn.grid(column=0, row=7, sticky="w", padx=3, pady=3)
        self.duration_lbl.grid(column=1, row=7, sticky="w", padx=3, pady=3)
        self.duration_min_lbl.grid(column=2, row=7, sticky="w", padx=3, pady=3)
        self.duration_min_entry.grid(column=3, row=7, sticky="nes", padx=3, pady=3)
        self.duration_max_lbl.grid(column=4, row=7, sticky="w", padx=3, pady=3)
        self.duration_max_entry.grid(column=5, row=7, sticky="nes", padx=3, pady=3)
        self.duration_disable_cbox.grid(column=6, row=7, sticky="nes", padx=3, pady=3)

        self.size_help_btn.grid(column=0, row=8, sticky="w", padx=3, pady=3)
        self.size_lbl.grid(column=1, row=8, sticky="w", padx=3, pady=3)
        self.img_size_max_lbl.grid(column=2, row=8, sticky="w", padx=3, pady=3)
        self.img_size_max_entry.grid(column=3, row=8, sticky="nes", padx=3, pady=3)
        self.vid_size_max_lbl.grid(column=4, row=8, sticky="w", padx=3, pady=3)
        self.vid_size_max_entry.grid(column=5, row=8, sticky="nes", padx=3, pady=3)
        self.size_disable_cbox.grid(column=6, row=8, sticky="nes", padx=3, pady=3)

        self.format_help_btn.grid(column=0, row=9, sticky="w", padx=3, pady=3)
        self.format_lbl.grid(column=1, row=9, sticky="w", padx=3, pady=3)
        self.img_format_lbl.grid(column=2, row=9, sticky="w", padx=3, pady=3)
        self.img_format_entry.grid(column=3, row=9, sticky="nes", padx=3, pady=3)
        self.vid_format_lbl.grid(column=4, row=9, sticky="w", padx=3, pady=3)
        self.vid_format_entry.grid(column=5, row=9, sticky="nes", padx=3, pady=3)

        self.power_help_btn1.grid(column=0, row=10, sticky="w", padx=3, pady=3)
        self.power_lbl1.grid(column=1, row=10, sticky="w", padx=3, pady=3)
        self.fps_power_lbl.grid(column=2, row=10, sticky="w", padx=3, pady=3)
        self.fps_power_entry.grid(column=3, row=10, sticky="w", padx=3, pady=3)
        self.res_power_lbl.grid(column=4, row=10, sticky="w", padx=3, pady=3)
        self.res_power_entry.grid(column=5, row=10, sticky="w", padx=3, pady=3)

        self.power_help_btn2.grid(column=0, row=11, sticky="w", padx=3, pady=3)
        self.power_lbl2.grid(column=1, row=11, sticky="w", padx=3, pady=3)
        self.quality_power_lbl.grid(column=2, row=11, sticky="w", padx=3, pady=3)
        self.quality_power_entry.grid(column=3, row=11, sticky="w", padx=3, pady=3)
        self.color_power_lbl.grid(column=4, row=11, sticky="w", padx=3, pady=3)
        self.color_power_entry.grid(column=5, row=11, sticky="w", padx=3, pady=3)

        self.fake_vid_help_btn.grid(column=0, row=12, sticky="w", padx=3, pady=3)
        self.fake_vid_lbl.grid(column=1, row=12, sticky="w", padx=3, pady=3)
        self.fake_vid_cbox.grid(column=6, row=12, sticky="nes", padx=3, pady=3)

        self.scale_filter_help_btn.grid(column=0, row=13, sticky="w", padx=3, pady=3)
        self.scale_filter_lbl.grid(column=1, row=13, sticky="w", padx=3, pady=3)
        self.scale_filter_opt.grid(
            column=2, row=13, columnspan=4, sticky="nes", padx=3, pady=3
        )

        self.quantize_method_help_btn.grid(column=0, row=14, sticky="w", padx=3, pady=3)
        self.quantize_method_lbl.grid(column=1, row=14, sticky="w", padx=3, pady=3)
        self.quantize_method_opt.grid(
            column=2, row=14, columnspan=4, sticky="nes", padx=3, pady=3
        )

        self.cache_dir_help_btn.grid(column=0, row=15, sticky="w", padx=3, pady=3)
        self.cache_dir_lbl.grid(column=1, row=15, sticky="w", padx=3, pady=3)
        self.cache_dir_entry.grid(
            column=2, row=15, columnspan=4, sticky="nes", padx=3, pady=3
        )

        self.default_emoji_help_btn.grid(column=0, row=16, sticky="w", padx=3, pady=3)
        self.default_emoji_lbl.grid(column=1, row=16, sticky="w", padx=3, pady=3)
        self.default_emoji_dsp.grid(column=6, row=16, sticky="nes", padx=3, pady=3)

        # https://stackoverflow.com/questions/43731784/tkinter-canvas-scrollbar-with-grid
        # Create a frame for the canvas with non-zero row&column weights
        self.frame_emoji_canvas.grid_rowconfigure(0, weight=1)
        self.frame_emoji_canvas.grid_columnconfigure(0, weight=1)
        # Set grid_propagate to False to allow buttons resizing later
        self.frame_emoji_canvas.grid_propagate(False)

        self.frame_advcomp.grid(column=0, row=0, sticky="news", padx=3, pady=3)
        self.frame_emoji_search.grid(column=0, row=1, sticky="news", padx=3, pady=3)
        self.frame_emoji_canvas.grid(column=0, row=2, sticky="news", padx=3, pady=3)

        self.categories_lbl = Label(
            self.frame_emoji_search,
            text="Category",
            width=15,
            justify="left",
            anchor="w",
        )
        self.categories_var = StringVar(self.scrollable_frame)
        self.categories_var.set("Smileys & Emotion")
        self.categories_opt = OptionMenu(
            self.frame_emoji_search,
            self.categories_var,
            "Smileys & Emotion",
            *self.categories,
            command=self.render_emoji_list,
            bootstyle="secondary",  # type: ignore
        )
        self.categories_opt.config(width=30)

        self.search_lbl = Label(self.frame_emoji_search, text="Search")
        self.search_var = StringVar(self.frame_emoji_search)
        self.search_var.trace_add("write", self.render_emoji_list)
        self.search_entry = Entry(self.frame_emoji_search, textvariable=self.search_var)
        self.search_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)

        self.categories_lbl.grid(column=0, row=0, sticky="nsw", padx=3, pady=3)
        self.categories_opt.grid(column=1, row=0, sticky="news", padx=3, pady=3)
        self.search_lbl.grid(column=0, row=1, sticky="nsw", padx=3, pady=3)
        self.search_entry.grid(column=1, row=1, sticky="news", padx=3, pady=3)

        # Add a canvas in frame_emoji_canvas
        self.emoji_canvas = Canvas(self.frame_emoji_canvas)
        self.emoji_canvas.grid(row=0, column=0, sticky="news")

        # Link a scrollbar to the canvas
        self.vsb = Scrollbar(
            self.frame_emoji_canvas,
            orient="vertical",
            command=self.emoji_canvas.yview,  # type: ignore
        )
        self.vsb.grid(row=0, column=1, sticky="ns")
        self.emoji_canvas.configure(yscrollcommand=self.vsb.set)

        self.frame_buttons = Frame(self.emoji_canvas)
        self.emoji_canvas.create_window((0, 0), window=self.frame_buttons, anchor="nw")

        self.render_emoji_list()

        self.set_emoji_btn()
        self.cb_disable_fps()
        self.cb_disable_res_w()
        self.cb_disable_res_h()
        self.cb_disable_quality()
        self.cb_disable_color()
        self.cb_disable_duration()
        self.cb_disable_size()
        self.cb_disable_format()
        self.cb_disable_fake_vid()

        GUIUtils.finalize_window(self)

    def cb_disable_fps(self, *_: Any) -> None:
        if self.gui.fps_disable_var.get() is True:
            state = "disabled"
        else:
            state = "normal"

        self.fps_min_entry.config(state=state)
        self.fps_max_entry.config(state=state)

    def cb_disable_res_w(self, *_: Any) -> None:
        if self.gui.res_w_disable_var.get() is True:
            state = "disabled"
        else:
            state = "normal"

        self.res_w_min_entry.config(state=state)
        self.res_w_max_entry.config(state=state)

    def cb_disable_res_h(self, *_: Any) -> None:
        if self.gui.res_h_disable_var.get() is True:
            state = "disabled"
        else:
            state = "normal"

        self.res_h_min_entry.config(state=state)
        self.res_h_max_entry.config(state=state)

    def cb_disable_quality(self, *_: Any) -> None:
        if self.gui.quality_disable_var.get() is True:
            state = "disabled"
        else:
            state = "normal"

        self.quality_min_entry.config(state=state)
        self.quality_max_entry.config(state=state)

    def cb_disable_color(self, *_: Any) -> None:
        if self.gui.color_disable_var.get() is True:
            state = "disabled"
        else:
            state = "normal"

        self.color_min_entry.config(state=state)
        self.color_max_entry.config(state=state)

    def cb_disable_duration(self, *_: Any) -> None:
        if (
            self.gui.duration_disable_var.get() is True
            or self.gui.comp_preset_var.get() == "auto"
        ):
            state = "disabled"
        else:
            state = "normal"

        self.duration_min_entry.config(state=state)
        self.duration_max_entry.config(state=state)

    def cb_disable_size(self, *_: Any) -> None:
        if (
            self.gui.size_disable_var.get() is True
            or self.gui.comp_preset_var.get() == "auto"
        ):
            state = "disabled"
        else:
            state = "normal"

        self.img_size_max_entry.config(state=state)
        self.vid_size_max_entry.config(state=state)

    def cb_disable_format(self, *_: Any) -> None:
        if self.gui.comp_preset_var.get() == "auto":
            state = "disabled"
        else:
            state = "normal"

        self.img_format_entry.config(state=state)
        self.vid_format_entry.config(state=state)

    def cb_disable_fake_vid(self, *_: Any) -> None:
        if self.gui.comp_preset_var.get() == "auto":
            state = "disabled"
        else:
            state = "normal"

        self.fake_vid_cbox.config(state=state)

    def set_emoji_btn(self) -> None:
        self.im = Image.new("RGBA", (128, 128), (255, 255, 255, 0))
        ImageDraw.Draw(self.im).text(  # type: ignore
            (0, 0),
            self.gui.default_emoji_var.get(),
            embedded_color=True,
            font=self.gui.emoji_font,
        )  # type: ignore
        self.im = self.im.resize((32, 32))
        self.ph_im = ImageTk.PhotoImage(self.im)
        self.default_emoji_dsp.config(image=self.ph_im)

    def render_emoji_list(self, *_: Any) -> None:
        category = self.categories_var.get()

        for emoji_btn, ph_im in self.emoji_btns:
            emoji_btn.destroy()
            del ph_im

        column = 0
        row = 0

        self.emoji_btns = []
        for entry in self.gui.emoji_list:
            # Filtering
            search_term = self.search_var.get().lower()
            emoji = entry["emoji"]
            keywords = entry["aliases"] + entry["tags"] + [emoji]
            if search_term == "":
                if entry["category"] != category:
                    continue
            else:
                ok = False

                if search_term in keywords:
                    ok = True
                elif len(search_term) >= 3:
                    for i in keywords:
                        if search_term in i:
                            ok = True

                if ok is False:
                    continue

            im = Image.new("RGBA", (196, 196), (255, 255, 255, 0))
            ImageDraw.Draw(im).text(  # type: ignore
                (16, 16), emoji, embedded_color=True, font=self.gui.emoji_font
            )  # type: ignore
            im = im.resize((32, 32))
            ph_im = ImageTk.PhotoImage(im)

            button = Button(
                self.frame_buttons,
                command=lambda i=emoji: self.cb_set_emoji(i),  # type: ignore
                bootstyle="dark",  # type: ignore
            )
            button.config(image=ph_im)
            button.grid(column=column, row=row)

            column += 1

            if column == self.emoji_column_per_row:
                column = 0
                row += 1

            self.emoji_btns.append((button, ph_im))

        # Update buttons frames idle tasks to let tkinter calculate buttons sizes
        self.frame_buttons.update_idletasks()

        # Resize the canvas frame to show specified number of buttons and the scrollbar
        if len(self.emoji_btns) > 0:
            in_view_columns_width = (
                self.emoji_btns[0][0].winfo_width() * self.emoji_column_per_row
            )
            in_view_rows_height = (
                self.emoji_btns[0][0].winfo_height() * self.emoji_visible_rows
            )
            self.frame_emoji_canvas.config(
                width=in_view_columns_width + self.vsb.winfo_width(),
                height=in_view_rows_height,
            )

        # Set the canvas scrolling region
        self.emoji_canvas.config(scrollregion=self.emoji_canvas.bbox("all"))

        # https://stackoverflow.com/questions/17355902/tkinter-binding-mousewheel-to-scrollbar
        self.emoji_canvas.bind("<Enter>", self.cb_bound_to_mousewheel)
        self.emoji_canvas.bind("<Leave>", self.cb_unbound_to_mousewheel)

    def cb_bound_to_mousewheel(self, event: Any) -> None:
        for i in self.mousewheel:
            self.emoji_canvas.bind_all(i, self.cb_on_mousewheel)

    def cb_unbound_to_mousewheel(self, event: Any) -> None:
        for i in self.mousewheel:
            self.emoji_canvas.unbind_all(i)

    def cb_on_mousewheel(self, event: Any) -> None:
        self.emoji_canvas.yview_scroll(
            int(-1 * (event.delta / self.delta_divide)), "units"
        )  # type: ignore

    def cb_set_emoji(self, emoji: str) -> None:
        self.gui.default_emoji_var.set(emoji)
        self.set_emoji_btn()



#!/usr/bin/env python3
from functools import partial
from threading import Thread
from typing import Any

from ttkbootstrap import Button, Entry, Frame, Label, LabelFrame  # type: ignore

from sticker_convert.gui_components.frames.right_clicker import RightClicker
from sticker_convert.gui_components.gui_utils import GUIUtils
from sticker_convert.gui_components.windows.base_window import BaseWindow
from sticker_convert.utils.auth.get_kakao_auth import GetKakaoAuth


class KakaoGetAuthWindow(BaseWindow):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        self.title("Get Kakao auth_token")

        self.cb_msg_block_kakao = partial(self.gui.cb_msg_block, parent=self)
        self.cb_ask_str_kakao = partial(self.gui.cb_ask_str, parent=self)

        self.frame_login_info = LabelFrame(
            self.scrollable_frame, text="Kakao login info"
        )
        self.frame_login_btn = Frame(self.scrollable_frame)

        self.frame_login_info.grid(column=0, row=0, sticky="news", padx=3, pady=3)
        self.frame_login_btn.grid(column=0, row=1, sticky="news", padx=3, pady=3)

        # Login info frame
        self.explanation1_lbl = Label(
            self.frame_login_info,
            text="This will simulate login to Android Kakao app",
            justify="left",
            anchor="w",
        )
        self.explanation2_lbl = Label(
            self.frame_login_info,
            text="You will send / receive verification code via SMS",
            justify="left",
            anchor="w",
        )
        self.explanation3_lbl = Label(
            self.frame_login_info,
            text="You maybe logged out of existing device",
            justify="left",
            anchor="w",
        )

        self.kakao_username_help_btn = Button(
            self.frame_login_info,
            text="?",
            width=1,
            command=lambda: self.cb_msg_block_kakao(
                self.gui.help["cred"]["kakao_username"]
            ),
            bootstyle="secondary",  # type: ignore
        )
        self.kakao_username_lbl = Label(
            self.frame_login_info, text="Username", width=18, justify="left", anchor="w"
        )
        self.kakao_username_entry = Entry(
            self.frame_login_info, textvariable=self.gui.kakao_username_var, width=30
        )
        self.kakao_username_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)

        self.kakao_password_help_btn = Button(
            self.frame_login_info,
            text="?",
            width=1,
            command=lambda: self.cb_msg_block_kakao(
                self.gui.help["cred"]["kakao_password"]
            ),
            bootstyle="secondary",  # type: ignore
        )
        self.kakao_password_lbl = Label(
            self.frame_login_info, text="Password", justify="left", anchor="w"
        )
        self.kakao_password_entry = Entry(
            self.frame_login_info, textvariable=self.gui.kakao_password_var, width=30
        )
        self.kakao_password_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)

        self.kakao_country_code_help_btn = Button(
            self.frame_login_info,
            text="?",
            width=1,
            command=lambda: self.cb_msg_block_kakao(
                self.gui.help["cred"]["kakao_country_code"]
            ),
            bootstyle="secondary",  # type: ignore
        )
        self.kakao_country_code_lbl = Label(
            self.frame_login_info, text="Country code", justify="left", anchor="w"
        )
        self.kakao_country_code_entry = Entry(
            self.frame_login_info,
            textvariable=self.gui.kakao_country_code_var,
            width=30,
        )
        self.kakao_country_code_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)

        self.kakao_phone_number_help_btn = Button(
            self.frame_login_info,
            text="?",
            width=1,
            command=lambda: self.cb_msg_block_kakao(
                self.gui.help["cred"]["kakao_phone_number"]
            ),
            bootstyle="secondary",  # type: ignore
        )
        self.kakao_phone_number_lbl = Label(
            self.frame_login_info, text="Phone number", justify="left", anchor="w"
        )
        self.kakao_phone_number_entry = Entry(
            self.frame_login_info,
            textvariable=self.gui.kakao_phone_number_var,
            width=30,
        )
        self.kakao_phone_number_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)

        self.explanation1_lbl.grid(
            column=0, row=0, columnspan=3, sticky="w", padx=3, pady=3
        )
        self.explanation2_lbl.grid(
            column=0, row=1, columnspan=3, sticky="w", padx=3, pady=3
        )
        self.explanation3_lbl.grid(
            column=0, row=2, columnspan=3, sticky="w", padx=3, pady=3
        )

        self.kakao_username_help_btn.grid(column=0, row=3, sticky="w", padx=3, pady=3)
        self.kakao_username_lbl.grid(column=1, row=3, sticky="w", padx=3, pady=3)
        self.kakao_username_entry.grid(column=2, row=3, sticky="w", padx=3, pady=3)

        self.kakao_password_help_btn.grid(column=0, row=4, sticky="w", padx=3, pady=3)
        self.kakao_password_lbl.grid(column=1, row=4, sticky="w", padx=3, pady=3)
        self.kakao_password_entry.grid(column=2, row=4, sticky="w", padx=3, pady=3)

        self.kakao_country_code_help_btn.grid(
            column=0, row=5, sticky="w", padx=3, pady=3
        )
        self.kakao_country_code_lbl.grid(column=1, row=5, sticky="w", padx=3, pady=3)
        self.kakao_country_code_entry.grid(column=2, row=5, sticky="w", padx=3, pady=3)

        self.kakao_phone_number_help_btn.grid(
            column=0, row=6, sticky="w", padx=3, pady=3
        )
        self.kakao_phone_number_lbl.grid(column=1, row=6, sticky="w", padx=3, pady=3)
        self.kakao_phone_number_entry.grid(column=2, row=6, sticky="w", padx=3, pady=3)

        # Login button frame
        self.login_btn = Button(
            self.frame_login_btn, text="Login and get auth_token", command=self.cb_login
        )

        self.login_btn.pack()

        GUIUtils.finalize_window(self)

    def cb_login(self) -> None:
        Thread(target=self.cb_login_thread, daemon=True).start()

    def cb_login_thread(self, *_: Any) -> None:
        self.gui.save_creds()
        m = GetKakaoAuth(
            opt_cred=self.gui.get_opt_cred(),
            cb_msg=self.gui.cb_msg,
            cb_msg_block=self.cb_msg_block_kakao,
            cb_ask_str=self.cb_ask_str_kakao,
        )

        auth_token = m.get_cred()

        if auth_token:
            if not self.gui.creds.get("kakao"):
                self.gui.creds["kakao"] = {}
            self.gui.creds["kakao"]["auth_token"] = auth_token
            self.gui.kakao_auth_token_var.set(auth_token)

            self.cb_msg_block_kakao(f"Got auth_token successfully: {auth_token}")
            self.gui.save_creds()
            self.gui.highlight_fields()
        else:
            self.cb_msg_block_kakao("Failed to get auth_token")



#!/usr/bin/env python3
import os
from pathlib import Path
from typing import Optional, Tuple, Union

from sticker_convert.job_option import CompOption
from sticker_convert.utils.media.codec_info import CodecInfo


class FormatVerify:
    @staticmethod
    def check_file(file: Union[Path, bytes], spec: CompOption) -> bool:
        if FormatVerify.check_presence(file) is False:
            return False

        file_info = CodecInfo(file)

        return (
            FormatVerify.check_file_res(
                file, res=spec.get_res(), square=spec.square, file_info=file_info
            )
            and FormatVerify.check_file_fps(
                file, fps=spec.get_fps(), file_info=file_info
            )
            and FormatVerify.check_file_duration(
                file, duration=spec.get_duration(), file_info=file_info
            )
            and FormatVerify.check_file_size(
                file, size=spec.get_size_max(), file_info=file_info
            )
            and FormatVerify.check_animated(
                file, animated=spec.animated, file_info=file_info
            )
            and FormatVerify.check_format(
                file, fmt=spec.get_format(), file_info=file_info
            )
        )

    @staticmethod
    def check_presence(file: Union[Path, bytes]) -> bool:
        if isinstance(file, Path):
            return Path(file).is_file()
        return True

    @staticmethod
    def check_file_res(
        file: Union[Path, bytes],
        res: Tuple[
            Tuple[Optional[int], Optional[int]], Tuple[Optional[int], Optional[int]]
        ],
        square: Optional[bool] = None,
        file_info: Optional[CodecInfo] = None,
    ) -> bool:
        if file_info:
            file_width, file_height = file_info.res
        else:
            file_width, file_height = CodecInfo.get_file_res(file)

        if res:
            if res[0][0] and file_width < res[0][0]:
                return False
            if res[0][1] and file_width > res[0][1]:
                return False
            if res[1][0] and file_height < res[1][0]:
                return False
            if res[1][1] and file_height > res[1][1]:
                return False
        if square and file_height != file_width:
            return False

        return True

    @staticmethod
    def check_file_fps(
        file: Union[Path, bytes],
        fps: Tuple[Optional[int], Optional[int]],
        file_info: Optional[CodecInfo] = None,
    ) -> bool:
        if file_info:
            file_fps = file_info.fps
        else:
            file_fps = CodecInfo.get_file_fps(file)

        if fps and fps[0] and file_fps < fps[0]:
            return False
        if fps and fps[1] and file_fps > fps[1]:
            return False

        return True

    @staticmethod
    def check_file_duration(
        file: Union[Path, bytes],
        duration: Tuple[Optional[int], Optional[int]],
        file_info: Optional[CodecInfo] = None,
    ) -> bool:
        if file_info:
            file_duration = file_info.duration
        else:
            file_duration = CodecInfo.get_file_duration(file)

        if file_duration == 0:
            return True
        if duration and duration[0] and file_duration < duration[0]:
            return False
        if duration and duration[1] and file_duration > duration[1]:
            return False

        return True

    @staticmethod
    def check_file_size(
        file: Union[Path, bytes],
        size: Tuple[Optional[int], Optional[int]],
        file_info: Optional[CodecInfo] = None,
    ) -> bool:
        if isinstance(file, Path):
            file_size = os.path.getsize(file)
        else:
            file_size = len(file)

        if file_info:
            file_animated = file_info.is_animated
        else:
            file_animated = CodecInfo.is_anim(file)

        if (
            file_animated is True
            and size
            and size[1] is not None
            and file_size > size[1]
        ):
            return False
        if (
            file_animated is False
            and size
            and size[0] is not None
            and file_size > size[0]
        ):
            return False

        return True

    @staticmethod
    def check_animated(
        file: Union[Path, bytes],
        animated: Optional[bool] = None,
        file_info: Optional[CodecInfo] = None,
    ) -> bool:
        if file_info:
            file_animated = file_info.is_animated
        else:
            file_animated = CodecInfo.is_anim(file)

        if animated is not None and file_animated != animated:
            return False

        return True

    @staticmethod
    def check_format(
        file: Union[Path, bytes],
        fmt: Tuple[Tuple[str, ...], Tuple[str, ...]],
        file_info: Optional[CodecInfo] = None,
    ) -> bool:
        if file_info:
            file_animated = file_info.is_animated
            file_ext = file_info.file_ext
        else:
            file_animated = CodecInfo.is_anim(file)
            if isinstance(file, Path):
                file_ext = CodecInfo.get_file_ext(file)
            else:
                file_ext = "." + CodecInfo.get_file_codec(file)

        jpg_exts = (".jpg", ".jpeg")
        png_exts = (".png", ".apng")

        if file_animated:
            valid_fmt = fmt[1]
        else:
            valid_fmt = fmt[0]

        if len(valid_fmt) == 0:
            return True
        if file_ext in valid_fmt:
            return True
        if file_ext in jpg_exts and (".jpg" in valid_fmt or ".jpeg" in valid_fmt):
            return True
        if file_ext in png_exts and (".png" in valid_fmt or ".apng" in valid_fmt):
            return True
        return False



#!/usr/bin/env python3
import json
import os
import platform
import shutil
from pathlib import Path
from typing import Optional, Tuple

from sqlcipher3 import dbapi2 as sqlite3


class GetSignalAuth:
    def get_signal_desktop(self) -> Tuple[Optional[str], Optional[str]]:
        if platform.system() == "Windows":
            signal_bin_path_prod = os.path.expandvars(
                "%localappdata%/Programs/signal-desktop/Signal.exe"
            )
            signal_bin_path_beta = os.path.expandvars(
                "%localappdata%/Programs/signal-desktop-beta/Signal Beta.exe"
            )
            signal_user_data_dir_prod = os.path.abspath(
                os.path.expandvars("%appdata%/Signal")
            )
            signal_user_data_dir_beta = os.path.abspath(
                os.path.expandvars("%appdata%/Signal Beta")
            )
        elif platform.system() == "Darwin":
            signal_bin_path_prod = "/Applications/Signal.app/Contents/MacOS/Signal"
            signal_bin_path_beta = (
                "/Applications/Signal Beta.app/Contents/MacOS/Signal Beta"
            )
            signal_user_data_dir_prod = os.path.expanduser(
                "~/Library/Application Support/Signal"
            )
            signal_user_data_dir_beta = os.path.expanduser(
                "~/Library/Application Support/Signal Beta"
            )
        else:
            if not (signal_bin_path_prod := shutil.which("signal-desktop")):  # type: ignore
                signal_bin_path_prod = "signal-desktop"
            if not (signal_bin_path_beta := shutil.which("signal-desktop-beta")):  # type: ignore
                signal_bin_path_beta = "signal-desktop-beta"
            signal_user_data_dir_prod = os.path.expanduser("~/.config/Signal")
            signal_user_data_dir_beta = os.path.expanduser("~/.config/Signal Beta")

        if Path(signal_bin_path_prod).is_file():
            return signal_bin_path_prod, signal_user_data_dir_prod
        if Path(signal_bin_path_beta).is_file():
            return signal_bin_path_beta, signal_user_data_dir_beta

        return None, None

    def get_cred(
        self,
        signal_bin_path: Optional[str] = None,
        signal_user_data_dir: Optional[str] = None,
    ) -> Tuple[Optional[str], Optional[str], str]:
        if not (signal_bin_path and signal_user_data_dir):
            signal_bin_path, signal_user_data_dir = self.get_signal_desktop()

        if not (signal_bin_path and signal_user_data_dir):
            msg = "Signal Desktop not detected.\n"
            msg += "Download and install Signal Desktop,\n"
            msg += "then login to Signal Desktop and try again."

            return None, None, msg

        signal_config = Path(signal_user_data_dir, "config.json")

        if not signal_config.is_file():
            msg = "Signal Desktop installed,\n"
            msg += "but it's config file not found.\n"
            msg += "Please login to Signal Desktop and try again.\n"
            msg += "\n"
            msg += f"{signal_bin_path=}\n"
            msg += f"{signal_user_data_dir=}\n"
            return None, None, msg

        with open(signal_config, encoding="utf-8") as f:
            config = json.load(f)
        key = config.get("key")
        db_key = f"x'{key}'"

        signal_database = Path(signal_user_data_dir, "sql/db.sqlite")

        if not signal_database.is_file():
            msg = "Signal Desktop installed,\n"
            msg += "but database file not found.\n"
            msg += "Please login to Signal Desktop and try again.\n"
            msg += "\n"
            msg += f"{signal_bin_path=}\n"
            msg += f"{signal_user_data_dir=}\n"
            return None, None, msg

        db_conn = sqlite3.connect(signal_database.as_posix())
        db_cursor = db_conn.cursor()
        db_cursor.execute(f'PRAGMA key="{db_key}"')

        uuid_id = None
        result = db_cursor.execute("SELECT * FROM items WHERE id='uuid_id'").fetchone()
        if result:
            uuid_id = json.loads(result[1])["value"]

        password = None
        result = db_cursor.execute("SELECT * FROM items WHERE id='password'").fetchone()
        if result:
            password = json.loads(result[1])["value"]

        db_conn.close()

        if uuid_id and password:
            msg = "Got uuid and password successfully:\n"
            msg += f"{uuid_id=}\n"
            msg += f"{password=}"
            return uuid_id, password, msg

        msg = "Signal Desktop installed and Database found,\n"
        msg += "but uuid and password not found.\n"
        msg += "Please login to Signal Desktop and try again.\n"
        msg += "\n"
        msg += f"{signal_bin_path=}\n"
        msg += f"{signal_user_data_dir=}\n"
        return None, None, msg



#!/usr/bin/env python3
from __future__ import annotations

import json
import os
import string
import zipfile
from io import BytesIO
from pathlib import Path
from queue import Queue
from typing import Any, Dict, List, Optional, Tuple, Union
from urllib import parse

import requests
from bs4 import BeautifulSoup
from PIL import Image

from sticker_convert.downloaders.download_base import DownloadBase
from sticker_convert.job_option import CredOption
from sticker_convert.utils.auth.get_line_auth import GetLineAuth
from sticker_convert.utils.callback import Callback, CallbackReturn, CbQueueItemType
from sticker_convert.utils.files.metadata_handler import MetadataHandler
from sticker_convert.utils.media.apple_png_normalize import ApplePngNormalize

# Reference: https://github.com/doubleplusc/Line-sticker-downloader/blob/master/sticker_dl.py


class MetadataLine:
    @staticmethod
    def analyze_url(url: str) -> Optional[Tuple[str, str, bool]]:
        region = ""
        is_emoji = False
        if url.startswith("line://shop/detail/"):
            pack_id = url.replace("line://shop/detail/", "")
            if len(url) == 24 and all(c in string.hexdigits for c in url):
                is_emoji = True
        elif url.startswith("https://store.line.me/stickershop/product/"):
            pack_id = url.replace(
                "https://store.line.me/stickershop/product/", ""
            ).split("/")[0]
            region = url.replace(
                "https://store.line.me/stickershop/product/", ""
            ).split("/")[1]
        elif url.startswith("https://line.me/S/sticker"):
            url_parsed = parse.urlparse(url)
            pack_id = url.replace("https://line.me/S/sticker/", "").split("/")[0]
            region = parse.parse_qs(url_parsed.query)["lang"][0]
        elif url.startswith("https://store.line.me/officialaccount/event/sticker/"):
            pack_id = url.replace(
                "https://store.line.me/officialaccount/event/sticker/", ""
            ).split("/")[0]
            region = url.replace(
                "https://store.line.me/officialaccount/event/sticker/", ""
            ).split("/")[1]
        elif url.startswith("https://store.line.me/emojishop/product/"):
            pack_id = url.replace("https://store.line.me/emojishop/product/", "").split(
                "/"
            )[0]
            region = url.replace("https://store.line.me/emojishop/product/", "").split(
                "/"
            )[1]
            is_emoji = True
        elif url.startswith("https://line.me/S/emoji"):
            url_parsed = parse.urlparse(url)
            pack_id = parse.parse_qs(url_parsed.query)["id"][0]
            region = parse.parse_qs(url_parsed.query)["lang"][0]
            is_emoji = True
        elif len(url) == 24 and all(c in string.hexdigits for c in url):
            pack_id = url
            is_emoji = True
        elif url.isnumeric():
            pack_id = url
        else:
            return None

        return pack_id, region, is_emoji

    @staticmethod
    def get_metadata_sticon(
        pack_id: str, region: str
    ) -> Optional[Tuple[str, str, List[Dict[str, Any]], str, bool]]:
        pack_meta_r = requests.get(
            f"https://stickershop.line-scdn.net/sticonshop/v1/{pack_id}/sticon/iphone/meta.json"
        )

        if pack_meta_r.status_code == 200:
            pack_meta = json.loads(pack_meta_r.text)
        else:
            return None

        if region == "":
            region = "en"

        pack_store_page = requests.get(
            f"https://store.line.me/emojishop/product/{pack_id}/{region}"
        )

        if pack_store_page.status_code != 200:
            return None

        pack_store_page_soup = BeautifulSoup(pack_store_page.text, "html.parser")

        title_tag = pack_store_page_soup.find(class_="mdCMN38Item01Txt")  # type: ignore
        if title_tag:
            title = title_tag.text
        else:
            return None

        author_tag = pack_store_page_soup.find(class_="mdCMN38Item01Author")  # type: ignore
        if author_tag:
            author = author_tag.text
        else:
            return None

        files = pack_meta["orders"]

        resource_type = pack_meta.get("sticonResourceType")
        has_sound = False

        return title, author, files, resource_type, has_sound

    @staticmethod
    def get_metadata_stickers(
        pack_id: str, region: str
    ) -> Optional[Tuple[str, str, List[Dict[str, Any]], str, bool]]:
        pack_meta_r = requests.get(
            f"https://stickershop.line-scdn.net/stickershop/v1/product/{pack_id}/android/productInfo.meta"
        )

        if pack_meta_r.status_code == 200:
            pack_meta = json.loads(pack_meta_r.text)
        else:
            return None

        if region == "":
            if "en" in pack_meta["title"]:
                # Prefer en release
                region = "en"
            else:
                # If no en release, use whatever comes first
                region = pack_meta["title"].keys()[0]

        if region == "zh-Hant":
            region = "zh_TW"

        title = pack_meta["title"].get("en")
        if title is None:
            title = pack_meta["title"][region]

        author = pack_meta["author"].get("en")
        if author is None:
            author = pack_meta["author"][region]

        files = pack_meta["stickers"]

        resource_type = pack_meta.get("stickerResourceType")
        has_sound = pack_meta.get("hasSound")

        return title, author, files, resource_type, has_sound


class DownloadLine(DownloadBase):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.headers = {
            "referer": "https://store.line.me",
            "user-agent": "Android",
            "x-requested-with": "XMLHttpRequest",
        }
        self.cookies = self.load_cookies()
        self.sticker_text_dict: Dict[int, Any] = {}

    def load_cookies(self) -> Dict[str, str]:
        cookies: Dict[str, str] = {}
        if self.opt_cred and self.opt_cred.line_cookies:
            line_cookies = self.opt_cred.line_cookies

            try:
                line_cookies_dict = json.loads(line_cookies)
                for c in line_cookies_dict:
                    cookies[c["name"]] = c["value"]
            except json.decoder.JSONDecodeError:
                try:
                    for i in line_cookies.split(";"):
                        c_key, c_value = i.split("=")
                        cookies[c_key] = c_value
                except ValueError:
                    self.cb.put(
                        'Warning: Line cookies invalid, you will not be able to add text to "Custom stickers"'
                    )

            if not GetLineAuth.validate_cookies(cookies):
                self.cb.put(
                    'Warning: Line cookies invalid, you will not be able to add text to "Custom stickers"'
                )
                cookies = {}

        return cookies

    def get_pack_url(self) -> str:
        # Reference: https://sora.vercel.app/line-sticker-download
        if self.is_emoji:
            if self.resource_type == "ANIMATION":
                pack_url = f"https://stickershop.line-scdn.net/sticonshop/v1/{self.pack_id}/sticon/iphone/package_animation.zip"
            else:
                pack_url = f"https://stickershop.line-scdn.net/sticonshop/v1/{self.pack_id}/sticon/iphone/package.zip"
        else:
            if (
                self.resource_type in ("ANIMATION", "ANIMATION_SOUND", "POPUP")
                or self.has_sound is True
            ):
                pack_url = f"https://stickershop.line-scdn.net/stickershop/v1/product/{self.pack_id}/iphone/stickerpack@2x.zip"
            elif self.resource_type == "PER_STICKER_TEXT":
                pack_url = f"https://stickershop.line-scdn.net/stickershop/v1/product/{self.pack_id}/iphone/sticker_custom_plus_base@2x.zip"
            elif self.resource_type == "NAME_TEXT":
                pack_url = f"https://stickershop.line-scdn.net/stickershop/v1/product/{self.pack_id}/iphone/sticker_name_base@2x.zip"
            else:
                pack_url = f"https://stickershop.line-scdn.net/stickershop/v1/product/{self.pack_id}/iphone/stickers@2x.zip"

        return pack_url

    def decompress(
        self,
        zf: zipfile.ZipFile,
        f_path: str,
        num: int,
        prefix: str = "",
        suffix: str = "",
    ) -> None:
        data = zf.read(f_path)
        ext = Path(f_path).suffix
        if ext == ".png" and not self.is_emoji and int() < 775:
            data = ApplePngNormalize.normalize(data)
        self.cb.put(f"Read {f_path}")

        out_path = Path(self.out_dir, prefix + str(num).zfill(3) + suffix + ext)
        with open(out_path, "wb") as f:
            f.write(data)

    def decompress_emoticon(self, zip_file: bytes) -> None:
        with zipfile.ZipFile(BytesIO(zip_file)) as zf:
            self.cb.put("Unzipping...")

            self.cb.put(
                (
                    "bar",
                    None,
                    {"set_progress_mode": "determinate", "steps": len(self.pack_files)},
                )
            )

            for num, sticker in enumerate(self.pack_files):
                if self.resource_type == "ANIMATION":
                    f_path = str(sticker) + "_animation.png"
                else:
                    f_path = str(sticker) + ".png"
                self.decompress(zf, f_path, num)

                self.cb.put("update_bar")

    def decompress_stickers(self, zip_file: bytes) -> None:
        with zipfile.ZipFile(BytesIO(zip_file)) as zf:
            self.cb.put("Unzipping...")

            self.cb.put(
                (
                    "bar",
                    None,
                    {"set_progress_mode": "determinate", "steps": len(self.pack_files)},
                )
            )

            for num, sticker in enumerate(self.pack_files):
                if self.resource_type in ("ANIMATION", "ANIMATION_SOUND"):
                    f_path = "animation@2x/" + str(sticker["id"]) + "@2x.png"
                elif self.resource_type == "POPUP":
                    if sticker.get("popup", {}).get("layer") == "BACKGROUND":
                        f_path = str(sticker["id"]) + "@2x.png"
                        self.decompress(zf, f_path, num, "preview-")
                    f_path = "popup/" + str(sticker["id"]) + ".png"
                else:
                    f_path = str(sticker["id"]) + "@2x.png"
                self.decompress(zf, f_path, num)

                if self.resource_type == "PER_STICKER_TEXT":
                    self.sticker_text_dict[num] = {
                        "sticker_id": sticker["id"],
                        "sticker_text": sticker["customPlus"]["defaultText"],
                    }

                elif self.resource_type == "NAME_TEXT":
                    self.sticker_text_dict[num] = {
                        "sticker_id": sticker["id"],
                        "sticker_text": "",
                    }

                if self.has_sound:
                    f_path = "sound/" + str(sticker["id"]) + ".m4a"
                    self.decompress(zf, f_path, num)

                self.cb.put("update_bar")

    def edit_custom_sticker_text(self) -> None:
        line_sticker_text_path = Path(self.out_dir, "line-sticker-text.txt")

        if not line_sticker_text_path.is_file():
            with open(line_sticker_text_path, "w+", encoding="utf-8") as f:
                json.dump(self.sticker_text_dict, f, indent=4, ensure_ascii=False)

            msg_block = (
                "The Line sticker pack you are downloading can have customized text.\n"
            )
            msg_block += "line-sticker-text.txt has been created in input directory.\n"
            msg_block += "Please edit line-sticker-text.txt, then continue."
            self.cb.put(("msg_block", (msg_block,), None))
            if self.cb_return:
                self.cb_return.get_response()

        with open(line_sticker_text_path, "r", encoding="utf-8") as f:
            self.sticker_text_dict = json.load(f)

    def get_custom_sticker_text_urls(self) -> List[Tuple[str, Path]]:
        custom_sticker_text_urls: List[Tuple[str, Path]] = []
        name_text_key_cache: Dict[str, str] = {}

        for num, data in self.sticker_text_dict.items():
            out_path = Path(self.out_dir, str(num).zfill(3))
            sticker_id = data["sticker_id"]
            sticker_text = data["sticker_text"]

            if self.resource_type == "PER_STICKER_TEXT":
                out_path_text = out_path.with_name(out_path.name + "-text.png")
                custom_sticker_text_urls.append(
                    (
                        f"https://store.line.me/overlay/sticker/{self.pack_id}/{sticker_id}/iPhone/sticker.png?text={parse.quote(sticker_text)}",
                        out_path_text,
                    )
                )

            elif self.resource_type == "NAME_TEXT" and sticker_text:
                out_path_text = out_path.with_name(out_path.name + "-text.png")
                name_text_key = name_text_key_cache.get(sticker_text, None)
                if not name_text_key:
                    name_text_key = self.get_name_text_key(sticker_text)
                if name_text_key:
                    name_text_key_cache[sticker_text] = name_text_key
                else:
                    continue

                custom_sticker_text_urls.append(
                    (
                        f"https://stickershop.line-scdn.net/stickershop/v1/sticker/{sticker_id}/iPhone/overlay/name/{name_text_key}/sticker@2x.png",
                        out_path_text,
                    )
                )

        return custom_sticker_text_urls

    def get_name_text_key(self, sticker_text: str) -> Optional[str]:
        params = {"text": sticker_text}

        response = requests.get(
            f"https://store.line.me/api/custom-sticker/preview/{self.pack_id}/{self.region}",
            params=params,
            cookies=self.cookies,
            headers=self.headers,
        )

        response_dict = json.loads(response.text)

        if response_dict["errorMessage"]:
            self.cb.put(
                f"Failed to generate customized text {sticker_text} due to: {response_dict['errorMessage']}"
            )
            return None

        name_text_key = (
            response_dict["productPayload"]["customOverlayUrl"]
            .split("name/")[-1]
            .split("/main.png")[0]
        )

        return name_text_key

    def combine_custom_text(self) -> None:
        for i in sorted(self.out_dir.iterdir()):
            if i.name.endswith("-text.png"):
                base_path = Path(self.out_dir, i.name.replace("-text.png", ".png"))
                text_path = Path(self.out_dir, i.name)

                with Image.open(base_path) as im:
                    base_img: Image.Image = im.convert("RGBA")

                with Image.open(text_path) as im:
                    text_img = im.convert("RGBA")

                with Image.alpha_composite(base_img, text_img) as im:
                    im.save(base_path)

                os.remove(text_path)

                self.cb.put(f"Combined {i.name.replace('-text.png', '.png')}")

    def download_stickers_line(self) -> bool:
        url_data = MetadataLine.analyze_url(self.url)
        if url_data:
            self.pack_id, self.region, self.is_emoji = url_data
        else:
            self.cb.put("Download failed: Unsupported URL format")
            return False

        if self.is_emoji:
            metadata = MetadataLine.get_metadata_sticon(self.pack_id, self.region)
        else:
            metadata = MetadataLine.get_metadata_stickers(self.pack_id, self.region)

        if metadata:
            (
                self.title,
                self.author,
                self.pack_files,
                self.resource_type,
                self.has_sound,
            ) = metadata
        else:
            self.cb.put("Download failed: Failed to get metadata")
            return False

        MetadataHandler.set_metadata(self.out_dir, title=self.title, author=self.author)

        pack_url = self.get_pack_url()
        zip_file = self.download_file(pack_url)

        if zip_file:
            self.cb.put(f"Downloaded {pack_url}")
        else:
            self.cb.put(f"Cannot download {pack_url}")
            return False

        if self.is_emoji:
            self.decompress_emoticon(zip_file)
        else:
            self.decompress_stickers(zip_file)

        custom_sticker_text_urls: List[Tuple[str, Path]] = []
        if self.sticker_text_dict and (
            self.resource_type == "PER_STICKER_TEXT"
            or (self.resource_type == "NAME_TEXT" and self.cookies)
        ):
            self.edit_custom_sticker_text()
            custom_sticker_text_urls = self.get_custom_sticker_text_urls()
        elif self.resource_type == "NAME_TEXT" and not self.cookies:
            self.cb.put('Warning: Line "Custom stickers" is supplied as input')
            self.cb.put(
                "However, adding custom message requires Line cookies, and it is not supplied"
            )
            self.cb.put("Continuing without adding custom text to stickers")

        self.download_multiple_files(custom_sticker_text_urls, headers=self.headers)
        self.combine_custom_text()

        return True

    @staticmethod
    def start(
        url: str,
        out_dir: Path,
        opt_cred: Optional[CredOption],
        cb: "Union[Queue[CbQueueItemType], Callback]",
        cb_return: CallbackReturn,
    ) -> bool:
        downloader = DownloadLine(url, out_dir, opt_cred, cb, cb_return)
        return downloader.download_stickers_line()



#!/usr/bin/env python3
from __future__ import annotations

import platform
from typing import TYPE_CHECKING, Tuple, Union

from ttkbootstrap import Canvas, Frame, PhotoImage, Scrollbar  # type: ignore

from sticker_convert.definitions import ROOT_DIR

if TYPE_CHECKING:
    from sticker_convert.gui import GUI
    from sticker_convert.gui_components.windows.base_window import BaseWindow


class GUIUtils:
    @staticmethod
    def set_icon(window: Union["BaseWindow", "GUI"]) -> None:
        window.icon = PhotoImage(file=ROOT_DIR / "resources/appicon.png")  # type: ignore
        window.iconphoto(1, window.icon)  # type: ignore
        if platform.system() == "Darwin":
            window.iconbitmap(bitmap=ROOT_DIR / "resources/appicon.icns")  # type: ignore
        elif platform.system() == "Windows":
            window.iconbitmap(bitmap=ROOT_DIR / "resources/appicon.ico")  # type: ignore
        window.tk.call("wm", "iconphoto", window._w, window.icon)  # type: ignore

    @staticmethod
    def create_scrollable_frame(
        window: Union["BaseWindow", "GUI"],
    ) -> Tuple[Frame, Frame, Canvas, Scrollbar, Scrollbar, Frame]:
        main_frame = Frame(window)
        main_frame.pack(fill="both", expand=1)

        horizontal_scrollbar_frame = Frame(main_frame)
        horizontal_scrollbar_frame.pack(fill="x", side="bottom")

        canvas = Canvas(main_frame)
        canvas.pack(side="left", fill="both", expand=1)

        x_scrollbar = Scrollbar(
            horizontal_scrollbar_frame,
            orient="horizontal",
            command=canvas.xview,  # type: ignore
        )
        x_scrollbar.pack(side="bottom", fill="x")

        y_scrollbar = Scrollbar(main_frame, orient="vertical", command=canvas.yview)  # type: ignore
        y_scrollbar.pack(side="right", fill="y")

        canvas.configure(xscrollcommand=x_scrollbar.set)
        canvas.configure(yscrollcommand=y_scrollbar.set)
        canvas.bind(
            "<Configure>", lambda e: canvas.config(scrollregion=canvas.bbox("all"))
        )

        scrollable_frame = Frame(canvas)
        canvas.create_window((0, 0), window=scrollable_frame, anchor="nw")

        return (
            main_frame,
            horizontal_scrollbar_frame,
            canvas,
            x_scrollbar,
            y_scrollbar,
            scrollable_frame,
        )

    @staticmethod
    def finalize_window(window: Union["GUI", "BaseWindow"]) -> None:
        window.attributes("-alpha", 0)  # type: ignore

        screen_width = window.winfo_screenwidth()
        screen_height = window.winfo_screenwidth()

        window.scrollable_frame.update_idletasks()

        frame_width = window.scrollable_frame.winfo_width()
        frame_height = window.scrollable_frame.winfo_height()

        window_width = frame_width + window.y_scrollbar.winfo_width()
        window_height = frame_height + window.x_scrollbar.winfo_height()

        window_width = min(window_width, screen_width)
        window_height = min(window_height, screen_height)

        frame_width = window_width - window.y_scrollbar.winfo_width()
        frame_height = window_height - window.x_scrollbar.winfo_height()

        window.maxsize(width=window_width, height=window_height)
        window.canvas.configure(width=frame_width, height=frame_height)

        window.place_window_center()

        window.attributes("-alpha", 1)  # type: ignore

        window.focus_force()



#!/usr/bin/env python3
import json
from pathlib import Path
from typing import Any, Dict


class JsonManager:
    @staticmethod
    def load_json(path: Path) -> Dict[Any, Any]:
        if not path.is_file():
            raise RuntimeError(f"{path} cannot be found")
        with open(path, encoding="utf-8") as f:
            data = json.load(f)
        return data

    @staticmethod
    def save_json(path: Path, data: Dict[Any, Any]) -> None:
        with open(path, "w+", encoding="utf-8") as f:
            json.dump(data, f, indent=4)



#!/usr/bin/env python3
import os
from decimal import ROUND_HALF_UP, Decimal
from fractions import Fraction
from io import BytesIO
from math import ceil, floor
from pathlib import Path
from queue import Queue
from typing import TYPE_CHECKING, Any, List, Literal, Optional, Tuple, Union, cast

import numpy as np
from PIL import Image

from sticker_convert.job_option import CompOption
from sticker_convert.utils.callback import Callback, CallbackReturn, CbQueueItemType
from sticker_convert.utils.files.cache_store import CacheStore
from sticker_convert.utils.media.codec_info import CodecInfo
from sticker_convert.utils.media.format_verify import FormatVerify

if TYPE_CHECKING:
    from av.video.plane import VideoPlane  # type: ignore

MSG_START_COMP = "[I] Start compressing {} -> {}"
MSG_SKIP_COMP = "[S] Compatible file found, skip compress and just copy {} -> {}"
MSG_COMP = (
    "[C] Compressing {} -> {} res={}x{}, "
    "quality={}, fps={}, color={} (step {}-{}-{})"
)
MSG_REDO_COMP = "[{}] Compressed {} -> {} but size {} {} limit {}, recompressing"
MSG_DONE_COMP = "[S] Successful compression {} -> {} size {} (step {})"
MSG_FAIL_COMP = (
    "[F] Failed Compression {} -> {}, "
    "cannot get below limit {} with lowest quality under current settings (Best size: {})"
)


def rounding(value: float) -> Decimal:
    return Decimal(value).quantize(0, ROUND_HALF_UP)


def get_step_value(
    max_step: Optional[int],
    min_step: Optional[int],
    step: int,
    steps: int,
    power: float = 1.0,
    even: bool = False,
) -> Optional[int]:
    # Power should be between -1 and positive infinity
    # Smaller power = More 'importance' of the parameter
    # Power of 1 is linear relationship
    # e.g. fps has lower power -> Try not to reduce it early on

    if step > 0:
        factor = pow(step / steps, power)
    else:
        factor = 0

    if max_step is not None and min_step is not None:
        v = round((max_step - min_step) * step / steps * factor + min_step)
        if even is True and v % 2 == 1:
            return v + 1
        return v
    return None


def useful_array(
    plane: "VideoPlane", bytes_per_pixel: int = 1, dtype: str = "uint8"
) -> "np.ndarray[Any, Any]":
    total_line_size = abs(plane.line_size)
    useful_line_size = plane.width * bytes_per_pixel
    arr: "np.ndarray[Any, Any]" = np.frombuffer(cast(bytes, plane), np.uint8)
    if total_line_size != useful_line_size:
        arr = arr.reshape(-1, total_line_size)[:, 0:useful_line_size].reshape(-1)
    return arr.view(np.dtype(dtype))


class StickerConvert:
    def __init__(
        self,
        in_f: Union[Path, Tuple[Path, bytes]],
        out_f: Path,
        opt_comp: CompOption,
        cb: "Union[Queue[CbQueueItemType], Callback]",
        #  cb_return: CallbackReturn
    ) -> None:
        self.in_f: Union[bytes, Path]
        if isinstance(in_f, Path):
            self.in_f = in_f
            self.in_f_name = self.in_f.name
            self.in_f_path = in_f
            self.codec_info_orig = CodecInfo(self.in_f)
        else:
            self.in_f = in_f[1]
            self.in_f_name = Path(in_f[0]).name
            self.in_f_path = in_f[0]
            self.codec_info_orig = CodecInfo(in_f[1], Path(in_f[0]).suffix)

        valid_formats: List[str] = []
        for i in opt_comp.get_format():
            valid_formats.extend(i)

        valid_ext = False
        self.out_f = Path()
        if len(valid_formats) == 0 or Path(out_f).suffix in valid_formats:
            self.out_f = Path(out_f)
            valid_ext = True

        if not valid_ext:
            if self.codec_info_orig.is_animated or opt_comp.fake_vid:
                ext = opt_comp.format_vid[0]
            else:
                ext = opt_comp.format_img[0]
            self.out_f = out_f.with_suffix(ext)

        self.out_f_name: str = self.out_f.name

        self.cb = cb
        self.frames_raw: "List[np.ndarray[Any, Any]]" = []
        self.frames_processed: "List[np.ndarray[Any, Any]]" = []
        self.opt_comp: CompOption = opt_comp
        if not self.opt_comp.steps:
            self.opt_comp.steps = 1

        self.size: int = 0
        self.size_max: Optional[int] = None
        self.res_w: Optional[int] = None
        self.res_h: Optional[int] = None
        self.quality: Optional[int] = None
        self.fps: Optional[Fraction] = None
        self.color: Optional[int] = None

        self.tmp_f: BytesIO = BytesIO()
        self.result: Optional[bytes] = None
        self.result_size: int = 0
        self.result_step: Optional[int] = None

        self.apngasm = None

    @staticmethod
    def convert(
        in_f: Union[Path, Tuple[Path, bytes]],
        out_f: Path,
        opt_comp: CompOption,
        cb: "Union[Queue[CbQueueItemType], Callback]",
        _cb_return: CallbackReturn,
    ) -> Tuple[bool, Path, Union[None, bytes, Path], int]:
        sticker = StickerConvert(in_f, out_f, opt_comp, cb)
        result = sticker._convert()
        cb.put("update_bar")
        return result

    def _convert(self) -> Tuple[bool, Path, Union[None, bytes, Path], int]:
        result = self.check_if_compatible()
        if result:
            return self.compress_done(result)

        self.cb.put((MSG_START_COMP.format(self.in_f_name, self.out_f_name)))

        steps_list = self.generate_steps_list()

        step_lower = 0
        step_upper = self.opt_comp.steps

        if self.codec_info_orig.is_animated is True:
            self.size_max = self.opt_comp.size_max_vid
        else:
            self.size_max = self.opt_comp.size_max_img

        if self.size_max is None:
            # No limit to size, create the best quality result
            step_current = 0
        else:
            step_current = int(rounding((step_lower + step_upper) / 2))

        self.frames_import()
        while True:
            param = steps_list[step_current]
            self.res_w = param[0]
            self.res_h = param[1]
            self.quality = param[2]
            if param[3] and self.codec_info_orig.fps:
                fps_tmp = min(param[3], self.codec_info_orig.fps)
                self.fps = self.fix_fps(fps_tmp)
            else:
                self.fps = Fraction(0)
            self.color = param[4]

            self.tmp_f = BytesIO()
            msg = MSG_COMP.format(
                self.in_f_name,
                self.out_f_name,
                self.res_w,
                self.res_h,
                self.quality,
                int(self.fps),
                self.color,
                step_lower,
                step_current,
                step_upper,
            )
            self.cb.put(msg)

            self.frames_processed = self.frames_drop(self.frames_raw)
            self.frames_processed = self.frames_resize(self.frames_processed)
            self.frames_export()

            self.tmp_f.seek(0)
            self.size = self.tmp_f.getbuffer().nbytes

            if not self.size_max or (
                self.size <= self.size_max and self.size >= self.result_size
            ):
                self.result = self.tmp_f.read()
                self.result_size = self.size
                self.result_step = step_current

            if step_upper - step_lower > 1 and self.size_max:
                if self.size <= self.size_max:
                    sign = "<"
                    step_upper = step_current
                else:
                    sign = ">"
                    step_lower = step_current
                step_current = int(rounding((step_lower + step_upper) / 2))
                self.recompress(sign)
            elif self.result:
                return self.compress_done(self.result, self.result_step)
            else:
                return self.compress_fail()

    def check_if_compatible(self) -> Optional[bytes]:
        if (
            FormatVerify.check_format(
                self.in_f,
                fmt=self.opt_comp.get_format(),
                file_info=self.codec_info_orig,
            )
            and FormatVerify.check_file_res(
                self.in_f, res=self.opt_comp.get_res(), file_info=self.codec_info_orig
            )
            and FormatVerify.check_file_fps(
                self.in_f, fps=self.opt_comp.get_fps(), file_info=self.codec_info_orig
            )
            and FormatVerify.check_file_size(
                self.in_f,
                size=self.opt_comp.get_size_max(),
                file_info=self.codec_info_orig,
            )
            and FormatVerify.check_file_duration(
                self.in_f,
                duration=self.opt_comp.get_duration(),
                file_info=self.codec_info_orig,
            )
        ):
            self.cb.put((MSG_SKIP_COMP.format(self.in_f_name, self.out_f_name)))

            if isinstance(self.in_f, Path):
                with open(self.in_f, "rb") as f:
                    result = f.read()
                self.result_size = os.path.getsize(self.in_f)
            else:
                result = self.in_f
                self.result_size = len(self.in_f)

            return result

        return None

    def generate_steps_list(self) -> List[Tuple[Optional[int], ...]]:
        steps_list: List[Tuple[Optional[int], ...]] = []
        for step in range(self.opt_comp.steps, -1, -1):
            steps_list.append(
                (
                    get_step_value(
                        self.opt_comp.res_w_max,
                        self.opt_comp.res_w_min,
                        step,
                        self.opt_comp.steps,
                        self.opt_comp.res_power,
                        True,
                    ),
                    get_step_value(
                        self.opt_comp.res_h_max,
                        self.opt_comp.res_h_min,
                        step,
                        self.opt_comp.steps,
                        self.opt_comp.res_power,
                        True,
                    ),
                    get_step_value(
                        self.opt_comp.quality_max,
                        self.opt_comp.quality_min,
                        step,
                        self.opt_comp.steps,
                        self.opt_comp.quality_power,
                    ),
                    get_step_value(
                        self.opt_comp.fps_max,
                        self.opt_comp.fps_min,
                        step,
                        self.opt_comp.steps,
                        self.opt_comp.fps_power,
                    ),
                    get_step_value(
                        self.opt_comp.color_max,
                        self.opt_comp.color_min,
                        step,
                        self.opt_comp.steps,
                        self.opt_comp.color_power,
                    ),
                )
            )

        return steps_list

    def recompress(self, sign: str) -> None:
        msg = MSG_REDO_COMP.format(
            sign, self.in_f_name, self.out_f_name, self.size, sign, self.size_max
        )
        self.cb.put(msg)

    def compress_fail(
        self,
    ) -> Tuple[bool, Path, Union[None, bytes, Path], int]:
        msg = MSG_FAIL_COMP.format(
            self.in_f_name, self.out_f_name, self.size_max, self.size
        )
        self.cb.put(msg)

        return False, self.in_f_path, self.out_f, self.size

    def compress_done(
        self, data: bytes, result_step: Optional[int] = None
    ) -> Tuple[bool, Path, Union[None, bytes, Path], int]:
        out_f: Union[None, bytes, Path]

        if self.out_f.stem == "none":
            out_f = None
        elif self.out_f.stem == "bytes":
            out_f = data
        else:
            out_f = self.out_f
            with open(self.out_f, "wb+") as f:
                f.write(data)

        if result_step:
            msg = MSG_DONE_COMP.format(
                self.in_f_name, self.out_f_name, self.result_size, result_step
            )
            self.cb.put(msg)

        return True, self.in_f_path, out_f, self.result_size

    def frames_import(self) -> None:
        if isinstance(self.in_f, Path):
            suffix = self.in_f.suffix
        else:
            suffix = Path(self.in_f_name).suffix

        if suffix in (".tgs", ".lottie", ".json"):
            self._frames_import_lottie()
        elif suffix in (".webp", ".apng", "png"):
            # ffmpeg do not support webp decoding (yet)
            # ffmpeg could fail to decode apng if file is buggy
            self._frames_import_pillow()
        else:
            self._frames_import_pyav()

    def _frames_import_pillow(self) -> None:
        with Image.open(self.in_f) as im:
            # Note: im.convert("RGBA") would return rgba image of current frame only
            if "n_frames" in dir(im):
                for i in range(im.n_frames):
                    im.seek(i)
                    self.frames_raw.append(np.asarray(im.convert("RGBA")))
            else:
                self.frames_raw.append(np.asarray(im.convert("RGBA")))

    def _frames_import_pyav(self) -> None:
        import av  # type: ignore
        from av.codec.context import CodecContext  # type: ignore
        from av.container.input import InputContainer  # type: ignore
        from av.video.codeccontext import VideoCodecContext  # type: ignore

        # Crashes when handling some webm in yuv420p and convert to rgba
        # https://github.com/PyAV-Org/PyAV/issues/1166
        file: Union[BytesIO, str]
        if isinstance(self.in_f, Path):
            file = self.in_f.as_posix()
        else:
            file = BytesIO(self.in_f)
        with av.open(file) as container:  # type: ignore
            container = cast(InputContainer, container)
            context = container.streams.video[0].codec_context
            if context.name == "vp8":
                context = CodecContext.create("libvpx", "r")  # type: ignore
            elif context.name == "vp9":
                context = CodecContext.create("libvpx-vp9", "r")  # type: ignore
            context = cast(VideoCodecContext, context)

            for packet in container.demux(container.streams.video):  # type: ignore
                for frame in context.decode(packet):
                    if frame.width % 2 != 0:
                        width = frame.width - 1
                    else:
                        width = frame.width
                    if frame.height % 2 != 0:
                        height = frame.height - 1
                    else:
                        height = frame.height
                    if frame.format.name == "yuv420p":
                        rgb_array = frame.to_ndarray(format="rgb24")  # type: ignore
                        cast("np.ndarray[Any, Any]", rgb_array)
                        rgba_array = np.dstack(
                            (
                                rgb_array,
                                np.zeros(rgb_array.shape[:2], dtype=np.uint8) + 255,
                            )  # type: ignore
                        )
                    else:
                        # yuva420p may cause crash
                        # https://github.com/laggykiller/sticker-convert/issues/114
                        frame = frame.reformat(  # type: ignore
                            width=width,
                            height=height,
                            format="yuva420p",
                            dst_colorspace=1,  # type: ignore
                        )

                        # https://stackoverflow.com/questions/72308308/converting-yuv-to-rgb-in-python-coefficients-work-with-array-dont-work-with-n
                        y = useful_array(frame.planes[0]).reshape(height, width)
                        u = useful_array(frame.planes[1]).reshape(
                            height // 2,
                            width // 2,
                        )
                        v = useful_array(frame.planes[2]).reshape(
                            height // 2,
                            width // 2,
                        )
                        a = useful_array(frame.planes[3]).reshape(height, width)

                        u = u.repeat(2, axis=0).repeat(2, axis=1)
                        v = v.repeat(2, axis=0).repeat(2, axis=1)

                        y = y.reshape((y.shape[0], y.shape[1], 1))
                        u = u.reshape((u.shape[0], u.shape[1], 1))
                        v = v.reshape((v.shape[0], v.shape[1], 1))
                        a = a.reshape((a.shape[0], a.shape[1], 1))

                        yuv_array = np.concatenate((y, u, v), axis=2)

                        yuv_array = yuv_array.astype(np.float32)
                        yuv_array[:, :, 0] = (
                            yuv_array[:, :, 0].clip(16, 235).astype(yuv_array.dtype)  # type: ignore
                            - 16
                        )
                        yuv_array[:, :, 1:] = (
                            yuv_array[:, :, 1:].clip(16, 240).astype(yuv_array.dtype)  # type: ignore
                            - 128
                        )

                        convert = np.array(
                            [
                                [1.164, 0.000, 1.793],
                                [1.164, -0.213, -0.533],
                                [1.164, 2.112, 0.000],
                            ]
                        )
                        rgb_array = (
                            np.matmul(yuv_array, convert.T).clip(0, 255).astype("uint8")
                        )
                        rgba_array = np.concatenate((rgb_array, a), axis=2)

                    self.frames_raw.append(rgba_array)

    def _frames_import_lottie(self) -> None:
        from rlottie_python.rlottie_wrapper import LottieAnimation

        if isinstance(self.in_f, Path):
            suffix = self.in_f.suffix
        else:
            suffix = Path(self.in_f_name).suffix

        if suffix == ".tgs":
            if isinstance(self.in_f, Path):
                anim = LottieAnimation.from_tgs(self.in_f.as_posix())
            else:
                import gzip

                with gzip.open(BytesIO(self.in_f)) as f:
                    data = f.read().decode(encoding="utf-8")
                anim = LottieAnimation.from_data(data)
        else:
            if isinstance(self.in_f, Path):
                anim = LottieAnimation.from_file(self.in_f.as_posix())
            else:
                anim = LottieAnimation.from_data(self.in_f.decode("utf-8"))

        for i in range(anim.lottie_animation_get_totalframe()):
            frame = np.asarray(anim.render_pillow_frame(frame_num=i))
            self.frames_raw.append(frame)

        anim.lottie_animation_destroy()

    def frames_resize(
        self, frames_in: "List[np.ndarray[Any, Any]]"
    ) -> "List[np.ndarray[Any, Any]]":
        frames_out: "List[np.ndarray[Any, Any]]" = []

        resample: Literal[0, 1, 2, 3, 4, 5]
        if self.opt_comp.scale_filter == "nearest":
            resample = Image.NEAREST
        elif self.opt_comp.scale_filter == "box":
            resample = Image.BOX
        elif self.opt_comp.scale_filter == "bilinear":
            resample = Image.BILINEAR
        elif self.opt_comp.scale_filter == "hamming":
            resample = Image.HAMMING
        elif self.opt_comp.scale_filter == "bicubic":
            resample = Image.BICUBIC
        elif self.opt_comp.scale_filter == "lanczos":
            resample = Image.LANCZOS
        else:
            resample = Image.LANCZOS

        for frame in frames_in:
            with Image.fromarray(frame, "RGBA") as im:  # type: ignore
                width, height = im.size

            if self.res_w is None:
                self.res_w = width
            if self.res_h is None:
                self.res_h = height

            if width > height:
                width_new = self.res_w
                height_new = height * self.res_w // width
            else:
                height_new = self.res_h
                width_new = width * self.res_h // height

            with im.resize((width_new, height_new), resample=resample) as im_resized:
                with Image.new(
                    "RGBA", (self.res_w, self.res_h), (0, 0, 0, 0)
                ) as im_new:
                    im_new.paste(
                        im_resized,
                        ((self.res_w - width_new) // 2, (self.res_h - height_new) // 2),
                    )
                    frames_out.append(np.asarray(im_new))

        return frames_out

    def frames_drop(
        self, frames_in: "List[np.ndarray[Any, Any]]"
    ) -> "List[np.ndarray[Any, Any]]":
        if (
            not self.codec_info_orig.is_animated
            or not self.fps
            or len(self.frames_processed) == 1
        ):
            return [frames_in[0]]

        frames_out: "List[np.ndarray[Any, Any]]" = []

        # fps_ratio: 1 frame in new anim equal to how many frame in old anim
        # speed_ratio: How much to speed up / slow down
        fps_ratio = self.codec_info_orig.fps / self.fps
        if (
            self.opt_comp.duration_min
            and self.codec_info_orig.duration < self.opt_comp.duration_min
        ):
            speed_ratio = self.codec_info_orig.duration / self.opt_comp.duration_min
        elif (
            self.opt_comp.duration_max
            and self.codec_info_orig.duration > self.opt_comp.duration_max
        ):
            speed_ratio = self.codec_info_orig.duration / self.opt_comp.duration_max
        else:
            speed_ratio = 1

        # How many frames to advance in original video for each frame of output video
        frame_increment = fps_ratio * speed_ratio

        frames_out_min = None
        frames_out_max = None
        if self.opt_comp.duration_min:
            frames_out_min = ceil(self.fps * self.opt_comp.duration_min / 1000)
        if self.opt_comp.duration_max:
            frames_out_max = floor(self.fps * self.opt_comp.duration_max / 1000)

        frame_current = 0
        frame_current_float = 0.0
        while True:
            frame_current_float += frame_increment
            frame_current = int(rounding(frame_current_float))
            if frame_current <= len(frames_in) - 1 and not (
                frames_out_max and len(frames_out) == frames_out_max
            ):
                frames_out.append(frames_in[frame_current])
            else:
                while len(frames_out) == 0 or (
                    frames_out_min and len(frames_out) < frames_out_min
                ):
                    frames_out.append(frames_in[-1])
                return frames_out

    def frames_export(self) -> None:
        is_animated = len(self.frames_processed) > 1 and self.fps
        if self.out_f.suffix in (".apng", ".png"):
            if is_animated:
                self._frames_export_apng()
            else:
                self._frames_export_png()
        elif self.out_f.suffix == ".webp" and is_animated:
            self._frames_export_webp()
        elif self.out_f.suffix in (".webm", ".mp4", ".mkv") or is_animated:
            self._frames_export_pyav()
        else:
            self._frames_export_pil()

    def _frames_export_pil(self) -> None:
        with Image.fromarray(self.frames_processed[0]) as im:  # type: ignore
            im.save(
                self.tmp_f,
                format=self.out_f.suffix.replace(".", ""),
                quality=self.quality,
            )

    def _frames_export_pyav(self) -> None:
        import av  # type: ignore
        from av.container import OutputContainer  # type: ignore
        from av.video.stream import VideoStream  # type: ignore

        options = {}

        if isinstance(self.quality, int):
            # Seems not actually working
            options["quality"] = str(self.quality)
            options["lossless"] = "0"

        if self.out_f.suffix == ".gif":
            codec = "gif"
            pixel_format = "rgb8"
            options["loop"] = "0"
        elif self.out_f.suffix in (".apng", ".png"):
            codec = "apng"
            pixel_format = "rgba"
            options["plays"] = "0"
        elif self.out_f.suffix in (".webp", ".webm", ".mkv"):
            codec = "libvpx-vp9"
            pixel_format = "yuva420p"
            options["loop"] = "0"
        else:
            codec = "libvpx-vp9"
            pixel_format = "yuv420p"
            options["loop"] = "0"

        with av.open(  # type: ignore
            self.tmp_f, "w", format=self.out_f.suffix.replace(".", "")
        ) as output:
            output = cast(OutputContainer, output)  # type: ignore
            out_stream = output.add_stream(codec, rate=self.fps, options=options)  # type: ignore
            out_stream = cast(VideoStream, out_stream)
            assert isinstance(self.res_w, int) and isinstance(self.res_h, int)
            out_stream.width = self.res_w
            out_stream.height = self.res_h
            out_stream.pix_fmt = pixel_format

            for frame in self.frames_processed:
                av_frame = av.VideoFrame.from_ndarray(frame, format="rgba")  # type: ignore
                for packet in out_stream.encode(av_frame):  # type: ignore
                    output.mux(packet)  # type: ignore

            for packet in out_stream.encode():  # type: ignore
                output.mux(packet)  # type: ignore

    def _frames_export_webp(self) -> None:
        import webp  # type: ignore

        assert self.fps

        config = webp.WebPConfig.new(quality=self.quality)  # type: ignore
        enc = webp.WebPAnimEncoder.new(self.res_w, self.res_h)  # type: ignore
        timestamp_ms = 0
        for frame in self.frames_processed:
            pic = webp.WebPPicture.from_numpy(frame)  # type: ignore
            enc.encode_frame(pic, timestamp_ms, config=config)  # type: ignore
            timestamp_ms += int(1000 / self.fps)
        anim_data = enc.assemble(timestamp_ms)  # type: ignore
        self.tmp_f.write(anim_data.buffer())  # type: ignore

    def _frames_export_png(self) -> None:
        with Image.fromarray(self.frames_processed[0], "RGBA") as image:  # type: ignore
            image_quant = self.quantize(image)

        with BytesIO() as f:
            image_quant.save(f, format="png")
            f.seek(0)
            frame_optimized = self.optimize_png(f.read())
            self.tmp_f.write(frame_optimized)

    def _frames_export_apng(self) -> None:
        from apngasm_python._apngasm_python import APNGAsm, create_frame_from_rgba  # type: ignore

        assert self.fps
        assert self.res_h

        frames_concat = np.concatenate(self.frames_processed)
        with Image.fromarray(frames_concat, "RGBA") as image_concat:  # type: ignore
            image_quant = self.quantize(image_concat)

        if self.apngasm is None:
            self.apngasm = APNGAsm()  # type: ignore
        assert isinstance(self.apngasm, APNGAsm)

        delay_num = int(1000 / self.fps)
        for i in range(0, image_quant.height, self.res_h):
            with BytesIO() as f:
                crop_dimension = (0, i, image_quant.width, i + self.res_h)
                image_cropped = image_quant.crop(crop_dimension)
                image_cropped.save(f, format="png")
                f.seek(0)
                frame_optimized = self.optimize_png(f.read())
            with Image.open(BytesIO(frame_optimized)) as im:
                image_final = im.convert("RGBA")
            frame_final = create_frame_from_rgba(
                np.array(image_final),
                width=image_final.width,
                height=image_final.height,
                delay_num=delay_num,
                delay_den=1000,
            )
            self.apngasm.add_frame(frame_final)

        with CacheStore.get_cache_store(path=self.opt_comp.cache_dir) as tempdir:
            tmp_apng = Path(tempdir, f"out{self.out_f.suffix}")
            self.apngasm.assemble(tmp_apng.as_posix())

            with open(tmp_apng, "rb") as f:
                self.tmp_f.write(f.read())

        self.apngasm.reset()

    def optimize_png(self, image_bytes: bytes) -> bytes:
        import oxipng

        return oxipng.optimize_from_memory(
            image_bytes,
            level=4,
            fix_errors=True,
            filter=[oxipng.RowFilter.Brute],
            optimize_alpha=True,
            strip=oxipng.StripChunks.safe(),
        )

    def quantize(self, image: Image.Image) -> Image.Image:
        if not (self.color and self.color <= 256):
            return image.copy()
        if self.opt_comp.quantize_method == "imagequant":
            return self._quantize_by_imagequant(image)
        if self.opt_comp.quantize_method == "fastoctree":
            return self._quantize_by_fastoctree(image)

        return image

    def _quantize_by_imagequant(self, image: Image.Image) -> Image.Image:
        import imagequant  # type: ignore

        assert self.quality
        assert self.opt_comp.quality_min
        assert self.opt_comp.quality_max
        assert self.color

        dither = 1 - (self.quality - self.opt_comp.quality_min) / (
            self.opt_comp.quality_max - self.opt_comp.quality_min
        )
        image_quant = None
        for i in range(self.quality, 101, 5):
            try:
                image_quant = imagequant.quantize_pil_image(  # type: ignore
                    image,
                    dithering_level=dither,
                    max_colors=self.color,
                    min_quality=self.opt_comp.quality_min,
                    max_quality=i,
                )
                return image_quant
            except RuntimeError:
                pass

        return image

    def _quantize_by_fastoctree(self, image: Image.Image) -> Image.Image:
        assert self.color

        return image.quantize(colors=self.color, method=2)

    def fix_fps(self, fps: float) -> Fraction:
        # After rounding fps/duration during export,
        # Video duration may exceed limit.
        # Hence we need to 'fix' the fps
        if self.out_f.suffix == ".gif":
            # Quote from https://www.w3.org/Graphics/GIF/spec-gif89a.txt
            # vii) Delay Time - If not 0, this field specifies
            # the number of hundredths (1/100) of a second
            #
            # For GIF, we need to adjust fps such that delay is matching to hundreths of second
            return self._fix_fps_duration(fps, 100)
        if self.out_f.suffix in (".webp", ".apng", ".png"):
            return self._fix_fps_duration(fps, 1000)

        return self._fix_fps_pyav(fps)

    def _fix_fps_duration(self, fps: float, denominator: int) -> Fraction:
        delay = int(rounding(denominator / fps))
        fps_fraction = Fraction(denominator, delay)
        if self.opt_comp.fps_max and fps_fraction > self.opt_comp.fps_max:
            return Fraction(denominator, (delay + 1))
        if self.opt_comp.fps_min and fps_fraction < self.opt_comp.fps_min:
            return Fraction(denominator, (delay - 1))
        return fps_fraction

    def _fix_fps_pyav(self, fps: float) -> Fraction:
        return Fraction(rounding(fps))



#!/usr/bin/env python3
import webbrowser
from typing import TYPE_CHECKING, Any

from ttkbootstrap import Button, Entry, Label, LabelFrame  # type: ignore

from sticker_convert.gui_components.frames.right_clicker import RightClicker
from sticker_convert.gui_components.windows.kakao_get_auth_window import KakaoGetAuthWindow
from sticker_convert.gui_components.windows.line_get_auth_window import LineGetAuthWindow
from sticker_convert.gui_components.windows.signal_get_auth_window import SignalGetAuthWindow

if TYPE_CHECKING:
    from sticker_convert.gui import GUI  # type: ignore


class CredFrame(LabelFrame):
    def __init__(self, gui: "GUI", *args: Any, **kwargs: Any) -> None:
        self.gui = gui
        super().__init__(*args, **kwargs)

        self.grid_columnconfigure(1, weight=1)

        self.signal_uuid_lbl = Label(
            self, text="Signal uuid", width=18, justify="left", anchor="w"
        )
        self.signal_uuid_entry = Entry(
            self,
            textvariable=self.gui.signal_uuid_var,
            width=50,
            validate="focusout",
            validatecommand=self.gui.highlight_fields,
        )
        self.signal_uuid_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)

        self.signal_password_lbl = Label(
            self, text="Signal password", justify="left", anchor="w"
        )
        self.signal_password_entry = Entry(
            self,
            textvariable=self.gui.signal_password_var,
            width=50,
            validate="focusout",
            validatecommand=self.gui.highlight_fields,
        )
        self.signal_password_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)

        self.signal_get_auth_btn = Button(
            self,
            text="Generate",
            command=self.cb_signal_get_auth,
            bootstyle="secondary",  # type: ignore
        )

        self.telegram_token_lbl = Label(
            self, text="Telegram token", justify="left", anchor="w"
        )
        self.telegram_token_entry = Entry(
            self,
            textvariable=self.gui.telegram_token_var,
            width=50,
            validate="focusout",
            validatecommand=self.gui.highlight_fields,
        )
        self.telegram_token_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)

        self.telegram_userid_lbl = Label(
            self, text="Telegram user_id", justify="left", anchor="w"
        )
        self.telegram_userid_entry = Entry(
            self,
            textvariable=self.gui.telegram_userid_var,
            width=50,
            validate="focusout",
            validatecommand=self.gui.highlight_fields,
        )
        self.telegram_userid_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)

        self.kakao_auth_token_lbl = Label(
            self, text="Kakao auth_token", justify="left", anchor="w"
        )
        self.kakao_auth_token_entry = Entry(
            self, textvariable=self.gui.kakao_auth_token_var, width=35
        )
        self.kakao_auth_token_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)
        self.kakao_get_auth_btn = Button(
            self,
            text="Generate",
            command=self.cb_kakao_get_auth,
            bootstyle="secondary",  # type: ignore
        )

        self.line_cookies_lbl = Label(
            self, text="Line cookies", width=18, justify="left", anchor="w"
        )
        self.line_cookies_entry = Entry(
            self, textvariable=self.gui.line_cookies_var, width=35
        )
        self.line_cookies_entry.bind("<Button-3><ButtonRelease-3>", RightClicker)
        self.line_get_auth_btn = Button(
            self,
            text="Generate",
            command=self.cb_line_get_auth,
            bootstyle="secondary",  # type: ignore
        )

        self.help_btn = Button(
            self,
            text="Get help",
            command=self.cb_cred_help,
            bootstyle="secondary",  # type: ignore
        )

        self.signal_uuid_lbl.grid(column=0, row=0, sticky="w", padx=3, pady=3)
        self.signal_uuid_entry.grid(
            column=1, row=0, columnspan=2, sticky="w", padx=3, pady=3
        )
        self.signal_password_lbl.grid(column=0, row=1, sticky="w", padx=3, pady=3)
        self.signal_password_entry.grid(
            column=1, row=1, columnspan=2, sticky="w", padx=3, pady=3
        )
        self.signal_get_auth_btn.grid(column=2, row=2, sticky="e", padx=3, pady=3)
        self.telegram_token_lbl.grid(column=0, row=3, sticky="w", padx=3, pady=3)
        self.telegram_token_entry.grid(
            column=1, row=3, columnspan=2, sticky="w", padx=3, pady=3
        )
        self.telegram_userid_lbl.grid(column=0, row=4, sticky="w", padx=3, pady=3)
        self.telegram_userid_entry.grid(
            column=1, row=4, columnspan=2, sticky="w", padx=3, pady=3
        )
        self.kakao_auth_token_lbl.grid(column=0, row=5, sticky="w", padx=3, pady=3)
        self.kakao_auth_token_entry.grid(column=1, row=5, sticky="w", padx=3, pady=3)
        self.kakao_get_auth_btn.grid(column=2, row=5, sticky="e", padx=3, pady=3)
        self.line_cookies_lbl.grid(column=0, row=6, sticky="w", padx=3, pady=3)
        self.line_cookies_entry.grid(column=1, row=6, sticky="w", padx=3, pady=3)
        self.line_get_auth_btn.grid(column=2, row=6, sticky="e", padx=3, pady=3)
        self.help_btn.grid(column=2, row=8, sticky="e", padx=3, pady=3)

    def cb_cred_help(self, *_: Any) -> None:
        faq_site = "https://github.com/laggykiller/sticker-convert#faq"
        success = webbrowser.open(faq_site)
        if not success:
            self.gui.cb_ask_str("You can get help from:", initialvalue=faq_site)

    def cb_kakao_get_auth(self, *_: Any) -> None:
        KakaoGetAuthWindow(self.gui)

    def cb_signal_get_auth(self, *_: Any) -> None:
        SignalGetAuthWindow(self.gui)

    def cb_line_get_auth(self, *_: Any) -> None:
        LineGetAuthWindow(self.gui)

    def set_states(self, state: str) -> None:
        self.signal_uuid_entry.config(state=state)
        self.signal_password_entry.config(state=state)
        self.signal_get_auth_btn.config(state=state)
        self.telegram_token_entry.config(state=state)
        self.telegram_userid_entry.config(state=state)
        self.kakao_auth_token_entry.config(state=state)
        self.kakao_get_auth_btn.config(state=state)
        self.line_cookies_entry.config(state=state)
        self.line_get_auth_btn.config(state=state)



#!/usr/bin/env python3
from __future__ import annotations

import json
from pathlib import Path
from typing import Dict, List, Optional, Tuple

from sticker_convert.utils.files.json_resources_loader import INPUT_JSON, OUTPUT_JSON
from sticker_convert.utils.media.codec_info import CodecInfo

RELATED_EXTENSIONS = (
    ".png",
    ".apng",
    ".jpg",
    ".jpeg",
    ".gif",
    ".tgs",
    ".lottie",
    ".json",
    ".mp4",
    ".mkv",
    ".mov",
    ".webm",
    ".webp",
    ".avi",
    ".m4a",
    ".wastickers",
)
RELATED_NAME = (
    "title.txt",
    "author.txt",
    "emoji.txt",
    "export-result.txt",
    ".DS_Store",
    "._.DS_Store",
)

BLACKLIST_PREFIX = ("cover",)
BLACKLIST_SUFFIX = (".txt", ".m4a", ".wastickers", ".DS_Store", "._.DS_Store")

XCODE_IMESSAGE_ICONSET = {
    "App-Store-1024x1024pt.png": (1024, 1024),
    "iPad-Settings-29pt@2x.png": (58, 58),
    "iPhone-settings-29pt@2x.png": (58, 58),
    "iPhone-settings-29pt@3x.png": (87, 87),
    "Messages27x20pt@2x.png": (54, 40),
    "Messages27x20pt@3x.png": (81, 60),
    "Messages32x24pt@2x.png": (64, 48),
    "Messages32x24pt@3x.png": (96, 72),
    "Messages-App-Store-1024x768pt.png": (1024, 768),
    "Messages-iPad-67x50pt@2x.png": (134, 100),
    "Messages-iPad-Pro-74x55pt@2x.png": (148, 110),
    "Messages-iPhone-60x45pt@2x.png": (120, 90),
    "Messages-iPhone-60x45pt@3x.png": (180, 135),
}


def check_if_xcodeproj(path: Path) -> bool:
    if not path.is_dir():
        return False
    for i in path.iterdir():
        if i.suffix == ".xcodeproj":
            return True
    return False


class MetadataHandler:
    @staticmethod
    def get_files_related_to_sticker_convert(
        directory: Path, include_archive: bool = True
    ) -> List[Path]:
        files = [
            i
            for i in sorted(directory.iterdir())
            if i.stem in RELATED_NAME
            or i.name in XCODE_IMESSAGE_ICONSET
            or i.suffix in RELATED_EXTENSIONS
            or (include_archive and i.name.startswith("archive_"))
            or check_if_xcodeproj(i)
        ]

        return files

    @staticmethod
    def get_stickers_present(directory: Path) -> List[Path]:
        stickers_present = [
            i
            for i in sorted(directory.iterdir())
            if Path(directory, i.name).is_file()
            and not i.name.startswith(BLACKLIST_PREFIX)
            and i.suffix not in BLACKLIST_SUFFIX
            and i.name not in XCODE_IMESSAGE_ICONSET
        ]

        return stickers_present

    @staticmethod
    def get_cover(directory: Path) -> Optional[Path]:
        stickers_present = sorted(directory.iterdir())
        for i in stickers_present:
            if Path(i).stem == "cover":
                return Path(directory, i.name)

        return None

    @staticmethod
    def get_metadata(
        directory: Path,
        title: Optional[str] = None,
        author: Optional[str] = None,
        emoji_dict: Optional[Dict[str, str]] = None,
    ) -> Tuple[Optional[str], Optional[str], Optional[Dict[str, str]]]:
        title_path = Path(directory, "title.txt")
        if not title and title_path.is_file():
            with open(title_path, encoding="utf-8") as f:
                title = f.read().strip()

        author_path = Path(directory, "author.txt")
        if not author and author_path.is_file():
            with open(author_path, encoding="utf-8") as f:
                author = f.read().strip()

        emoji_path = Path(directory, "emoji.txt")
        if not emoji_dict and emoji_path.is_file():
            with open(emoji_path, "r", encoding="utf-8") as f:
                emoji_dict = json.load(f)

        return title, author, emoji_dict

    @staticmethod
    def set_metadata(
        directory: Path,
        title: Optional[str] = None,
        author: Optional[str] = None,
        emoji_dict: Optional[Dict[str, str]] = None,
    ) -> None:
        title_path = Path(directory, "title.txt")
        if title is not None:
            with open(title_path, "w+", encoding="utf-8") as f:
                f.write(title)

        author_path = Path(directory, "author.txt")
        if author is not None:
            with open(author_path, "w+", encoding="utf-8") as f:
                f.write(author)

        emoji_path = Path(directory, "emoji.txt")
        if emoji_dict is not None:
            with open(emoji_path, "w+", encoding="utf-8") as f:
                json.dump(emoji_dict, f, indent=4, ensure_ascii=False)

    @staticmethod
    def check_metadata_provided(
        input_dir: Path, input_option: str, metadata: str
    ) -> bool:
        """
        Check if metadata provided via .txt file (if from local)
        or will be provided by input source (if not from local)
        Does not check if metadata provided via user input in GUI or flag options
        metadata = 'title' or 'author'
        """
        input_presets = INPUT_JSON
        assert input_presets

        if input_option == "local":
            metadata_file_path = Path(input_dir, f"{metadata}.txt")
            metadata_provided = metadata_file_path.is_file()
            if metadata_provided:
                with open(metadata_file_path, encoding="utf-8") as f:
                    metadata_provided = bool(f.read())
        else:
            metadata_provided = input_presets[input_option]["metadata_provides"][
                metadata
            ]

        return metadata_provided

    @staticmethod
    def check_metadata_required(output_option: str, metadata: str) -> bool:
        # metadata = 'title' or 'author'
        output_presets = OUTPUT_JSON
        assert output_presets
        return output_presets[output_option]["metadata_requirements"][metadata]

    @staticmethod
    def generate_emoji_file(directory: Path, default_emoji: str = "") -> None:
        emoji_path = Path(directory, "emoji.txt")
        emoji_dict = None
        if emoji_path.is_file():
            with open(emoji_path, "r", encoding="utf-8") as f:
                emoji_dict = json.load(f)

        emoji_dict_new = {}
        for file in sorted(directory.iterdir()):
            if not Path(directory, file).is_file() and CodecInfo.get_file_ext(file) in (
                ".txt",
                ".m4a",
            ):
                continue
            file_name = Path(file).stem
            if emoji_dict and file_name in emoji_dict:
                emoji_dict_new[file_name] = emoji_dict[file_name]
            else:
                emoji_dict_new[file_name] = default_emoji

        with open(emoji_path, "w+", encoding="utf-8") as f:
            json.dump(emoji_dict_new, f, indent=4, ensure_ascii=False)

    @staticmethod
    def split_sticker_packs(
        directory: Path,
        title: str,
        file_per_pack: Optional[int] = None,
        file_per_anim_pack: Optional[int] = None,
        file_per_image_pack: Optional[int] = None,
        separate_image_anim: bool = True,
    ) -> Dict[str, List[Path]]:
        # {pack_1: [sticker1_path, sticker2_path]}
        packs: Dict[str, List[Path]] = {}

        if file_per_pack is None:
            file_per_pack = (
                file_per_anim_pack
                if file_per_anim_pack is not None
                else file_per_image_pack
            )
        else:
            file_per_anim_pack = file_per_pack
            file_per_image_pack = file_per_pack

        stickers_present = MetadataHandler.get_stickers_present(directory)

        processed = 0

        if separate_image_anim is True:
            image_stickers: List[Path] = []
            anim_stickers: List[Path] = []

            image_pack_count = 0
            anim_pack_count = 0

            anim_present = False
            image_present = False

            for processed, file in enumerate(stickers_present):
                file_path = directory / file

                if CodecInfo.is_anim(file_path):
                    anim_stickers.append(file_path)
                else:
                    image_stickers.append(file_path)

                anim_present = anim_present or len(anim_stickers) > 0
                image_present = image_present or len(image_stickers) > 0

                finished_all = processed == len(stickers_present) - 1

                if len(anim_stickers) == file_per_anim_pack or (
                    finished_all and len(anim_stickers) > 0
                ):
                    suffix = f'{"-anim" if image_present else ""}{"-" + str(anim_pack_count) if anim_pack_count > 0 else ""}'
                    title_current = str(title) + suffix
                    packs[title_current] = anim_stickers.copy()
                    anim_stickers = []
                    anim_pack_count += 1
                if len(image_stickers) == file_per_image_pack or (
                    finished_all and len(image_stickers) > 0
                ):
                    suffix = f'{"-image" if anim_present else ""}{"-" + str(image_pack_count) if image_pack_count > 0 else ""}'
                    title_current = str(title) + suffix
                    packs[title_current] = image_stickers.copy()
                    image_stickers = []
                    image_pack_count += 1

        else:
            stickers: List[Path] = []
            pack_count = 0

            for processed, file in enumerate(stickers_present):
                file_path = Path(directory, file)

                stickers.append(file_path)

                finished_all = processed == len(stickers_present) - 1

                if len(stickers) == file_per_pack or (
                    finished_all and len(stickers) > 0
                ):
                    suffix = f'{"-" + str(pack_count) if pack_count > 0 else ""}'
                    title_current = str(title) + suffix
                    packs[title_current] = stickers.copy()
                    stickers = []
                    pack_count += 1

        return packs



#!/usr/bin/env python3
from queue import Queue
from typing import Union

from sticker_convert.job_option import CompOption, CredOption, OutputOption
from sticker_convert.utils.callback import Callback, CallbackReturn, CbQueueItemType


class UploadBase:
    def __init__(
        self,
        opt_output: OutputOption,
        opt_comp: CompOption,
        opt_cred: CredOption,
        cb: "Union[Queue[CbQueueItemType], Callback]",
        cb_return: CallbackReturn,
    ) -> None:
        if not cb:
            cb = Callback(silent=True)
            cb_return = CallbackReturn()

        self.opt_output = opt_output
        self.opt_comp = opt_comp
        self.opt_cred = opt_cred
        self.cb = cb
        self.cb_return = cb_return

        self.base_spec = CompOption(
            fps_power=self.opt_comp.fps_power,
            res_power=self.opt_comp.res_power,
            quality_min=self.opt_comp.quality_min,
            quality_max=self.opt_comp.quality_max,
            quality_power=self.opt_comp.quality_power,
            color_min=self.opt_comp.color_min,
            color_max=self.opt_comp.color_max,
            color_power=self.opt_comp.color_power,
            quantize_method=self.opt_comp.quantize_method,
            scale_filter=self.opt_comp.scale_filter,
            steps=self.opt_comp.steps,
            cache_dir=self.opt_comp.cache_dir,
        )



#!/usr/bin/env python3
from typing import TYPE_CHECKING, Any

from ttkbootstrap import Button, Frame  # type: ignore

if TYPE_CHECKING:
    from sticker_convert.gui import GUI


class ControlFrame(Frame):
    def __init__(self, gui: "GUI", *args: Any, **kwargs: Any) -> None:
        self.gui = gui
        super().__init__(*args, **kwargs)

        self.start_btn = Button(
            self,
            text="Start",
            command=self.cb_start_btn,
            bootstyle="default",  # type: ignore
        )

        self.start_btn.pack(expand=True, fill="x")

    def cb_start_btn(self, *args: Any, **kwargs: Any) -> None:
        if self.gui.job:
            response = self.gui.cb_ask_bool("Cancel job?")
            if response is True:
                self.gui.cancel_job()
        else:
            self.gui.start_job()



from typing import Dict

from sticker_convert.definitions import ROOT_DIR
from sticker_convert.utils.files.json_manager import JsonManager

HELP_JSON: Dict[str, Dict[str, str]] = JsonManager.load_json(
    ROOT_DIR / "resources/help.json"
)
INPUT_JSON = JsonManager.load_json(ROOT_DIR / "resources/input.json")
COMPRESSION_JSON = JsonManager.load_json(ROOT_DIR / "resources/compression.json")
OUTPUT_JSON = JsonManager.load_json(ROOT_DIR / "resources/output.json")
EMOJI_JSON = JsonManager.load_json(ROOT_DIR / "resources/emoji.json")



#!/usr/bin/env python3
import os
import platform
import shutil
from contextlib import contextmanager
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, ContextManager, Generator, Optional, Union
from uuid import uuid4

if platform.system() == "Linux":
    import memory_tempfile  # type: ignore

    tempfile = memory_tempfile.MemoryTempfile(fallback=True)  # type: ignore
else:
    import tempfile


def debug_cache_dir(path: str) -> ContextManager[Path]:
    @contextmanager
    def generator() -> Generator[Any, Any, Any]:
        path_random = Path(path, str(uuid4()))
        os.mkdir(path_random)
        try:
            yield path_random
        finally:
            shutil.rmtree(path_random)

    return generator()


class CacheStore:
    @staticmethod
    def get_cache_store(
        path: Optional[str] = None,
    ) -> "Union[ContextManager[Path], TemporaryDirectory[str]]":
        if path:
            return debug_cache_dir(path)
        return tempfile.TemporaryDirectory()  # type: ignore



import os
import sys
from pathlib import Path
from typing import List, Optional

import pytest
from _pytest._py.path import LocalPath

from tests.common import COMPRESSION_DICT, PYTHON_EXE, SAMPLE_DIR, SIGNAL_PASSWORD, SIGNAL_UUID, SRC_DIR, TELEGRAM_TOKEN, TELEGRAM_USERID, run_cmd

os.chdir(Path(__file__).resolve().parent)
sys.path.append("../src")

from sticker_convert.utils.media.codec_info import CodecInfo  # type: ignore # noqa: E402

TEST_UPLOAD = os.environ.get("TEST_UPLOAD")


def _run_sticker_convert(
    tmp_path: LocalPath, preset: str, export: Optional[str]
) -> None:
    preset_dict = COMPRESSION_DICT.get(preset)

    cmd: List[str] = [
        PYTHON_EXE,
        "sticker-convert.py",
        "--input-dir",
        str(SAMPLE_DIR),
        "--output-dir",
        str(tmp_path),
        "--preset",
        preset,
        "--no-confirm",
        "--author",
        "sticker-convert-test",
        "--title",
        "sticker-convert-test",
    ]

    if export:
        cmd.append(f"--export-{export}")

    run_cmd(cmd, cwd=SRC_DIR)

    for i in SAMPLE_DIR.iterdir():
        preset_dict.get("size_max").get("img")

        if i.name.startswith("static_") and preset_dict.get("fake_vid") is False:
            size_max = preset_dict.get("size_max").get("img")
            fmt: str = preset_dict.get("format").get("img")
        else:
            size_max = preset_dict.get("size_max").get("vid")
            fmt = preset_dict.get("format").get("vid")

        fname = i.stem + fmt
        fpath = Path(tmp_path) / fname
        fps, frames, duration = CodecInfo.get_file_fps_frames_duration(fpath)

        print(f"[TEST] Check if {fname} exists")
        assert fpath.is_file()
        assert os.path.getsize(fpath) < size_max

        if i.name.startswith("animated_"):
            print(f"[TEST] {fname}: {fps=} {frames=} {duration=}")
            duration_min = preset_dict.get("duration").get("min")
            duration_max = preset_dict.get("duration").get("max")
            assert fps <= preset_dict.get("fps").get("max")
            if duration_min:
                assert duration >= duration_min
            if duration_max:
                assert duration <= duration_max


def _xcode_asserts(tmp_path: LocalPath) -> None:
    iconset = {
        "App-Store-1024x1024pt.png": (1024, 1024),
        "iPad-Settings-29pt@2x.png": (58, 58),
        "iPhone-settings-29pt@2x.png": (58, 58),
        "iPhone-Settings-29pt@3x.png": (87, 87),
        "Messages27x20pt@2x.png": (54, 40),
        "Messages27x20pt@3x.png": (81, 60),
        "Messages32x24pt@2x.png": (64, 48),
        "Messages32x24pt@3x.png": (96, 72),
        "Messages-App-Store-1024x768pt.png": (1024, 768),
        "Messages-iPad-67x50pt@2x.png": (134, 100),
        "Messages-iPad-Pro-74x55pt@2x.png": (148, 110),
        "Messages-iPhone-60x45pt@2x.png": (120, 90),
        "Messages-iPhone-60x45pt@3x.png": (180, 135),
    }

    imessage_xcode_dir = Path(tmp_path) / "sticker-convert-test"

    assert Path(imessage_xcode_dir / "sticker-convert-test/Info.plist").is_file()

    assert Path(
        imessage_xcode_dir / "sticker-convert-test StickerPackExtension/Info.plist"
    ).is_file()
    assert Path(
        imessage_xcode_dir
        / "sticker-convert-test StickerPackExtension/Stickers.xcstickers/Contents.json"
    ).is_file()
    for i in iconset:
        assert Path(
            imessage_xcode_dir
            / "sticker-convert-test StickerPackExtension/Stickers.xcstickers/iMessage App Icon.stickersiconset"
            / i
        ).is_file()

    assert Path(
        imessage_xcode_dir / "sticker-convert-test.xcodeproj/project.pbxproj"
    ).is_file()


@pytest.mark.skipif(not TEST_UPLOAD, reason="TEST_UPLOAD not set")
@pytest.mark.skipif(not (SIGNAL_UUID and SIGNAL_PASSWORD), reason="No credentials")
def test_upload_signal_with_upload(tmp_path: LocalPath) -> None:
    _run_sticker_convert(tmp_path, "signal", "signal")


@pytest.mark.skipif(not TEST_UPLOAD, reason="TEST_UPLOAD not set")
@pytest.mark.skipif(not (TELEGRAM_TOKEN and TELEGRAM_USERID), reason="No credentials")
def test_upload_telegram_with_upload(tmp_path: LocalPath) -> None:
    _run_sticker_convert(tmp_path, "telegram", "telegram")


@pytest.mark.skipif(not TEST_UPLOAD, reason="TEST_UPLOAD not set")
@pytest.mark.skipif(not (TELEGRAM_TOKEN and TELEGRAM_USERID), reason="No credentials")
def test_upload_telegram_emoji_with_upload(tmp_path: LocalPath) -> None:
    _run_sticker_convert(tmp_path, "telegram_emoji", None)


@pytest.mark.skipif(
    TELEGRAM_TOKEN is not None and TELEGRAM_USERID is not None,
    reason="With credentials",
)
def test_upload_signal(tmp_path: LocalPath) -> None:
    _run_sticker_convert(tmp_path, "signal", None)


@pytest.mark.skipif(
    TELEGRAM_TOKEN is not None and TELEGRAM_USERID is not None,
    reason="With credentials",
)
def test_upload_telegram(tmp_path: LocalPath) -> None:
    _run_sticker_convert(tmp_path, "telegram", None)


@pytest.mark.skipif(
    TELEGRAM_TOKEN is not None and TELEGRAM_USERID is not None,
    reason="With credentials",
)
def test_upload_telegram_emoji(tmp_path: LocalPath) -> None:
    _run_sticker_convert(tmp_path, "telegram_emoji", None)


def test_export_wastickers(tmp_path: LocalPath) -> None:
    _run_sticker_convert(tmp_path, "whatsapp", "whatsapp")

    wastickers_path = Path(tmp_path, "sticker-convert-test.wastickers")
    assert Path(wastickers_path).is_file()


def test_export_line(tmp_path: LocalPath) -> None:
    _run_sticker_convert(tmp_path, "line", None)


def test_export_kakao(tmp_path: LocalPath) -> None:
    _run_sticker_convert(tmp_path, "kakao", None)


def test_export_xcode_imessage_small(tmp_path: LocalPath) -> None:
    _run_sticker_convert(tmp_path, "imessage_small", "imessage")

    _xcode_asserts(tmp_path)


def test_export_xcode_imessage_medium(tmp_path: LocalPath) -> None:
    _run_sticker_convert(tmp_path, "imessage_medium", "imessage")

    _xcode_asserts(tmp_path)


def test_export_xcode_imessage_large(tmp_path: LocalPath) -> None:
    _run_sticker_convert(tmp_path, "imessage_large", "imessage")

    _xcode_asserts(tmp_path)



#!/usr/bin/env python3


def main() -> None:
    import multiprocessing
    import sys

    from sticker_convert.version import __version__

    multiprocessing.freeze_support()
    print(f"sticker-convert {__version__}")
    if len(sys.argv) == 1:
        print("Launching GUI...")
        from sticker_convert.gui import GUI

        GUI().gui()
    else:
        from sticker_convert.cli import CLI

        CLI().cli()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import webbrowser
from functools import partial
from threading import Thread
from typing import Any

from ttkbootstrap import Button, Frame, Label  # type: ignore

from sticker_convert.gui_components.gui_utils import GUIUtils
from sticker_convert.gui_components.windows.base_window import BaseWindow
from sticker_convert.utils.auth.get_line_auth import GetLineAuth


class LineGetAuthWindow(BaseWindow):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        self.title("Get Line cookie")

        self.cb_msg_block_line = partial(self.gui.cb_msg_block, parent=self)

        self.frame_info = Frame(self.scrollable_frame)
        self.frame_btn = Frame(self.scrollable_frame)

        self.frame_info.grid(column=0, row=0, sticky="news", padx=3, pady=3)
        self.frame_btn.grid(column=0, row=1, sticky="news", padx=3, pady=3)

        # Info frame
        self.explanation1_lbl = Label(
            self.frame_info,
            text="Line cookies are required to create custom message stickers",
            justify="left",
            anchor="w",
        )
        self.explanation2_lbl = Label(
            self.frame_info,
            text="Please open web browser and login to Line",
            justify="left",
            anchor="w",
        )
        self.explanation3_lbl = Label(
            self.frame_info,
            text='After that, press "Get cookies"',
            justify="left",
            anchor="w",
        )

        self.explanation1_lbl.grid(
            column=0, row=0, columnspan=3, sticky="w", padx=3, pady=3
        )
        self.explanation2_lbl.grid(
            column=0, row=1, columnspan=3, sticky="w", padx=3, pady=3
        )
        self.explanation3_lbl.grid(
            column=0, row=2, columnspan=3, sticky="w", padx=3, pady=3
        )

        # Buttons frame
        self.open_browser_btn = Button(
            self.frame_btn, text="Open browser", command=self.cb_open_browser
        )
        self.get_cookies_btn = Button(
            self.frame_btn, text="Get cookies", command=self.cb_get_cookies
        )

        self.open_browser_btn.pack()
        self.get_cookies_btn.pack()

        GUIUtils.finalize_window(self)

    def cb_open_browser(self) -> None:
        line_login_site = "https://store.line.me/login"
        success = webbrowser.open(line_login_site)
        if not success:
            self.gui.cb_ask_str(
                "Cannot open web browser for you. Install web browser and open:",
                initialvalue=line_login_site,
            )

    def cb_get_cookies(self) -> None:
        Thread(target=self.cb_get_cookies_thread, daemon=True).start()

    def cb_get_cookies_thread(self, *_: Any) -> None:
        m = GetLineAuth()

        line_cookies = None
        line_cookies = m.get_cred()

        if line_cookies:
            if not self.gui.creds.get("line"):
                self.gui.creds["line"] = {}
            self.gui.creds["line"]["cookies"] = line_cookies
            self.gui.line_cookies_var.set(line_cookies)

            self.cb_msg_block_line("Got Line cookies successfully")
            self.gui.save_creds()
            self.gui.highlight_fields()
            return

        self.cb_msg_block_line(
            "Failed to get Line cookies. Have you logged in the web browser?"
        )
